from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import RosFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class RouterOSVendor(AbstractVendor):
    NAME = "routeros"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        # FIXME: пока не удалось победить \x1b[c после включения safe mode
        # if len(cmds) > 99:
        #     raise Exception("RouterOS does not support more 100 actions in safe mode")
        # before.add_cmd(RosDevice.SAFE_MODE)
        pass
        # after.add_cmd(RosDevice.SAFE_MODE)

        return before, after

    def match(self) -> list[str]:
        return ["RouterOS"]

    @property
    def reverse(self) -> str:
        return "remove"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("RouterOS")

    def make_formatter(self, **kwargs) -> RosFormatter:
        return RosFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return ""



from dataclasses import dataclass
from typing import Callable, Any, Sequence

from .match_args import MatchExpr, PairMatcher, SingleMatcher
from .match_args import MatchedArgs
from .device_models import GlobalOptionsDTO
from .peer_models import MeshSession, IndirectPeerDTO, VirtualLocalDTO, VirtualPeerDTO, DirectPeerDTO
from .port_processor import PortProcessor, united_ports
from .basemodel import is_dataclass_empty


class DirectPeer(DirectPeerDTO):
    match: MatchedArgs
    device: Any
    ports: list[str]
    all_connected_ports: list[str]

    def __init__(self, match: MatchedArgs, device: Any, ports: list[str], all_connected_ports: list[str]) -> None:
        super().__init__()
        self.match = match
        self.device = device
        self.ports = ports
        self.all_connected_ports = all_connected_ports

    def is_empty(self):
        if not is_dataclass_empty(self.family_options):
            return False
        return self.__dict__.keys() == {"match", "device", "ports", "all_connected_ports", "family_options"}


class IndirectPeer(IndirectPeerDTO):
    match: MatchedArgs
    device: Any

    def __init__(self, match: MatchedArgs, device: Any) -> None:
        super().__init__()
        self.match = match
        self.device = device

    def is_empty(self):
        if not is_dataclass_empty(self.family_options):
            return False
        return self.__dict__.keys() == {"match", "device", "family_options"}


class VirtualLocal(VirtualLocalDTO):
    match: MatchedArgs
    device: Any

    def __init__(self, match: MatchedArgs, device: Any) -> None:
        super().__init__()
        self.match = match
        self.device = device

    def is_empty(self):
        if not is_dataclass_empty(self.family_options):
            return False
        return self.__dict__.keys() == {"match", "device", "family_options"}


class VirtualPeer(VirtualPeerDTO):
    num: int

    def is_empty(self):
        return self.__dict__.keys() == {"num"}


class GlobalOptions(GlobalOptionsDTO):
    match: MatchedArgs
    device: Any

    def __init__(self, match: MatchedArgs, device: Any) -> None:
        super().__init__()
        self.match = match
        self.device = device

    def is_empty(self):
        return self.__dict__.keys() == {"match", "device"}


GlobalHandler = Callable[[GlobalOptions], None]


@dataclass
class GlobalRule:
    matcher: SingleMatcher
    handler: GlobalHandler


DirectHandler = Callable[[DirectPeer, DirectPeer, MeshSession], None]
IndirectHandler = Callable[[IndirectPeer, IndirectPeer, MeshSession], None]
VirtualHandler = Callable[[VirtualLocal, VirtualPeer, MeshSession], None]


@dataclass
class DirectRule:
    __slots__ = ("matcher", "handler", "port_processor")
    matcher: PairMatcher
    handler: DirectHandler
    port_processor: PortProcessor


@dataclass
class IndirectRule:
    __slots__ = ("matcher", "handler")
    matcher: PairMatcher
    handler: IndirectHandler


@dataclass
class VirtualRule:
    __slots__ = ("matcher", "num", "handler")
    matcher: SingleMatcher
    num: Sequence[int]
    handler: VirtualHandler


@dataclass
class MatchedGlobal:
    __slots__ = ("match", "handler")
    match: MatchedArgs
    handler: GlobalHandler


@dataclass
class MatchedDirectPair:
    __slots__ = ("handler", "port_processor", "direct_order", "name_left", "name_right", "match_left", "match_right")
    handler: DirectHandler
    port_processor: PortProcessor
    direct_order: bool
    name_left: str
    name_right: str
    match_left: MatchedArgs
    match_right: MatchedArgs


@dataclass
class MatchedIndirectPair:
    __slots__ = ("handler", "direct_order", "name_left", "name_right", "match_left", "match_right")
    handler: IndirectHandler
    direct_order: bool
    name_left: str
    name_right: str
    match_left: MatchedArgs
    match_right: MatchedArgs


@dataclass
class MatchedVirtualPair:
    __slots__ = ("match", "num", "handler")
    match: MatchedArgs
    num: Sequence[int]
    handler: VirtualHandler


class MeshRulesRegistry:
    def __init__(self, match_short_name: bool = False):
        self.direct_rules: list[DirectRule] = []
        self.indirect_rules: list[IndirectRule] = []
        self.global_rules: list[GlobalRule] = []
        self.virtual_rules: list[VirtualRule] = []
        self.nested: list[MeshRulesRegistry] = []
        self.match_short_name = match_short_name

    def _normalize_host(self, host: str) -> str:
        if self.match_short_name:
            return host.split(".", maxsplit=1)[0]
        return host

    def include(self, nested_registry: "MeshRulesRegistry") -> None:
        self.nested.append(nested_registry)

    def device(self, peer_mask: str, *match: MatchExpr) -> Callable[[GlobalHandler], GlobalHandler]:
        matcher = SingleMatcher(peer_mask, match)

        def register(handler: GlobalHandler) -> GlobalHandler:
            self.global_rules.append(GlobalRule(matcher, handler))
            return handler

        return register

    def direct(
            self, left_mask: str, right_mask: str, *match: MatchExpr,
            port_processor: PortProcessor = united_ports,
    ) -> Callable[[DirectHandler], DirectHandler]:
        matcher = PairMatcher(left_mask, right_mask, match)

        def register(handler: DirectHandler) -> DirectHandler:
            self.direct_rules.append(DirectRule(matcher, handler, port_processor))
            return handler

        return register

    def indirect(
            self, left_mask: str, right_mask: str, *match: MatchExpr,
    ) -> Callable[[IndirectHandler], IndirectHandler]:
        matcher = PairMatcher(left_mask, right_mask, match)

        def register(handler: IndirectHandler) -> IndirectHandler:
            self.indirect_rules.append(IndirectRule(matcher, handler))
            return handler

        return register

    def virtual(
            self, peer_mask: str, num: Sequence[int], *match: MatchExpr,
    ) -> Callable[[VirtualHandler], VirtualHandler]:
        matcher = SingleMatcher(peer_mask, match)

        def register(handler: VirtualHandler) -> VirtualHandler:
            self.virtual_rules.append(VirtualRule(matcher, num, handler))
            return handler

        return register

    def lookup_direct(self, device: str, neighbors: list[str]) -> list[MatchedDirectPair]:
        found = []
        device_norm = self._normalize_host(device)
        for neighbor in neighbors:
            neighbor_norm = self._normalize_host(neighbor)
            for rule in self.direct_rules:
                if args := rule.matcher.match_pair(device_norm, neighbor_norm):
                    found.append(MatchedDirectPair(
                        handler=rule.handler,
                        port_processor=rule.port_processor,
                        direct_order=True,
                        name_left=device,
                        name_right=neighbor,
                        match_left=args[0],
                        match_right=args[1],
                    ))
                if args := rule.matcher.match_pair(neighbor_norm, device_norm):
                    found.append(MatchedDirectPair(
                        handler=rule.handler,
                        port_processor=rule.port_processor,
                        direct_order=False,
                        name_left=neighbor,
                        name_right=device,
                        match_left=args[0],
                        match_right=args[1],
                    ))
        for registry in self.nested:
            found.extend(registry.lookup_direct(device, neighbors))
        return found

    def lookup_indirect(self, device: str, devices: list[str]) -> list[MatchedIndirectPair]:
        found = []
        device_norm = self._normalize_host(device)
        for other_device in devices:
            other_device_norm = self._normalize_host(other_device)
            for rule in self.indirect_rules:
                if args := rule.matcher.match_pair(device_norm, other_device_norm):
                    found.append(MatchedIndirectPair(
                        handler=rule.handler,
                        direct_order=True,
                        name_left=device,
                        name_right=other_device,
                        match_left=args[0],
                        match_right=args[1],
                    ))
                if args := rule.matcher.match_pair(other_device_norm, device_norm):
                    found.append(MatchedIndirectPair(
                        handler=rule.handler,
                        direct_order=False,
                        name_left=other_device,
                        name_right=device,
                        match_left=args[0],
                        match_right=args[1],
                    ))
        for registry in self.nested:
            found.extend(registry.lookup_indirect(device, devices))
        return found

    def lookup_virtual(self, device: str) -> list[MatchedVirtualPair]:
        found = []
        device_norm = self._normalize_host(device)
        for rule in self.virtual_rules:
            if args := rule.matcher.match_one(device_norm):
                found.append(MatchedVirtualPair(
                    handler=rule.handler,
                    match=args,
                    num=rule.num,
                ))
        for registry in self.nested:
            found.extend(registry.lookup_virtual(device))
        return found

    def lookup_global(self, device: str) -> list[MatchedGlobal]:
        found = []
        device_norm = self._normalize_host(device)
        for rule in self.global_rules:
            if args := rule.matcher.match_one(device_norm):
                found.append(MatchedGlobal(
                    handler=rule.handler,
                    match=args,
                ))
        for registry in self.nested:
            found.extend(registry.lookup_global(device))
        return found



from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Sequence, Collection, Iterator, Mapping
from typing import Any

from annet.generators import PartialGenerator
from annet.rpl import SingleCondition, MatchField, ThenField, RoutingPolicy, ConditionOperator
from .entities import (
    CommunityList, CommunityLogic, CommunityType, arista_well_known_community, mangle_united_community_list_name,
)


def get_used_community_lists(
        communities: Collection[CommunityList], policies: Collection[RoutingPolicy],
) -> list[CommunityList]:
    assert_unique_names(communities)
    communities_dict = {c.name: c for c in communities}
    used_communities: set[str] = set()
    for policy in policies:
        for statement in policy.statements:
            condition: SingleCondition[Sequence[str]]
            for match_field in (
                    MatchField.community, MatchField.large_community,
                    MatchField.extcommunity_rt, MatchField.extcommunity_soo
            ):
                for condition in statement.match.find_all(match_field):
                    used_communities.update(condition.value)
            for then_field in (
                    ThenField.community, ThenField.large_community,
                    ThenField.extcommunity_rt, ThenField.extcommunity_soo,
                    ThenField.extcommunity,
            ):
                for action in statement.then.find_all(then_field):
                    if action.value.replaced is not None:
                        used_communities.update(action.value.replaced)
                    used_communities.update(action.value.added)
                    used_communities.update(action.value.removed)
    return [
        communities_dict[name] for name in sorted(used_communities)
    ]


def get_used_united_community_lists(
        communities: Collection[CommunityList], policies: Collection[RoutingPolicy],
) -> list[list[CommunityList]]:
    """
    Return communities united into groups according to HAS_ANY policy
    """
    assert_unique_names(communities)
    communities_dict = {c.name: c for c in communities}
    used_communities: dict[str, list[CommunityList]] = {}
    for policy in policies:
        for statement in policy.statements:
            condition: SingleCondition[Sequence[str]]
            for match_field in (
                    MatchField.community, MatchField.large_community,
                    MatchField.extcommunity_rt, MatchField.extcommunity_soo
            ):
                for condition in statement.match.find_all(match_field):
                    if (
                            condition.operator == ConditionOperator.HAS_ANY and
                            len(condition.value) > 1
                    ):
                        united_name = mangle_united_community_list_name(condition.value)
                        united_communities: list[CommunityList] = [
                            communities_dict[name] for name in condition.value
                        ]
                        if not all(united_communities[0].type == c.type for c in united_communities):
                            raise ValueError(
                                f"Cannot apply HAS_ANY to communities of different types, "
                                f"found for policy: `{policy.name}`, statement: {statement.name}"
                            )
                        if not all(united_communities[0].use_regex == c.use_regex for c in united_communities):
                            raise ValueError(
                                f"Cannot apply HAS_ANY to communities with different use_regex flag, "
                                f"found for policy: `{policy.name}`, statement: {statement.name}"
                            )
                        used_communities[united_name] = united_communities
                    else:
                        for name in condition.value:
                            used_communities[name] = [communities_dict[name]]
            for then_field in (
                    ThenField.community, ThenField.large_community,
                    ThenField.extcommunity_rt, ThenField.extcommunity_soo
            ):
                for action in statement.then.find_all(then_field):
                    if action.value.replaced is not None:
                        for name in action.value.replaced:
                            used_communities[name] = [communities_dict[name]]
                    for name in action.value.added:
                        used_communities[name] = [communities_dict[name]]
                    for name in action.value.removed:
                        used_communities[name] = [communities_dict[name]]
    return [
        used_communities[name] for name in sorted(used_communities)
    ]


def assert_unique_names(communities: Collection[CommunityList]) -> None:
    duplicated: list[str] = []
    seen_names: set[str] = set()
    for c in communities:
        if c.name in seen_names:
            duplicated.append(c.name)
        seen_names.add(c.name)
    if duplicated:
        raise NotImplementedError(f"Non-unique community-list names are not supported: {duplicated}")


class CommunityListGenerator(PartialGenerator, ABC):
    TAGS = ["policy", "rpl", "routing"]

    @abstractmethod
    def get_policies(self, device: Any) -> list[RoutingPolicy]:
        raise NotImplementedError()

    @abstractmethod
    def get_community_lists(self, device: Any) -> list[CommunityList]:
        raise NotImplementedError()

    def get_used_community_lists(self, device: Any) -> list[CommunityList]:
        return get_used_community_lists(
            communities=self.get_community_lists(device),
            policies=self.get_policies(device),
        )

    def get_used_united_community_lists(self, device: Any) -> list[list[CommunityList]]:
        return get_used_united_community_lists(
            communities=self.get_community_lists(device),
            policies=self.get_policies(device),
        )

    def acl_huawei(self, _):
        return r"""
        ip community-filter
        ip extcommunity-filter
        ip extcommunity-list
        ip large-community-filter
        """

    def ref_huawei(self, _):
        return """
        route-policy
            if-match community-filter <name>
            if-match extcommunity-filter <name>
            if-match extcommunity-list soo <name>
            if-match large-community-filter <name>
            apply comm-filter <name>
        """

    def _huawei_community_filter(self, index: int, community_list: CommunityList, members: str) -> Sequence[str]:
        if community_list.use_regex:
            match_type = "advanced"
        else:
            match_type = "basic"
        if community_list.type is CommunityType.BASIC:
            return "ip community-filter", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.RT:
            return "ip extcommunity-filter", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.SOO:
            return "ip extcommunity-list soo", match_type, community_list.name, f"index {index}", "permit", members
        elif community_list.type is CommunityType.LARGE:
            return "ip large-community-filter", match_type, community_list.name, f"index {index}", "permit", members
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for huawei")

    def run_huawei(self, device: Any):
        for community_list in self.get_used_community_lists(device):
            if community_list.use_regex and len(community_list.members) > 1:
                raise NotImplementedError("Multiple regex is not supported for huawei")
            if community_list.type is CommunityType.RT:
                # RT communities used with prefix rt
                members: Sequence[str] = [f"rt {m}" for m in community_list.members]
            else:
                members = community_list.members

            if community_list.logic == CommunityLogic.AND:
                # to get AND logic the communities should be in one sting
                yield self._huawei_community_filter(10, community_list, " ".join(members))
            elif community_list.logic == CommunityLogic.OR:
                for i, member in enumerate(members):
                    member_id = (i + 1) * 10
                    yield self._huawei_community_filter(member_id, community_list, member)
            else:
                raise NotImplementedError(f"Community logic {community_list.logic} is not implemented for huawei")

    def acl_arista(self, _):
        return r"""
        ip community-list
        ip extcommunity-list
        """

    def _arista_community_list(
            self, name: str, use_regex: bool, comm_type: CommunityType, members: str,
    ) -> Sequence[str]:
        if use_regex:
            match_type = "regexp"
        else:
            match_type = ""
        if comm_type is CommunityType.BASIC:
            return "ip community-list", match_type, name, "permit", members
        elif comm_type is CommunityType.RT:
            return "ip extcommunity-list", match_type, name, "permit", members
        elif comm_type is CommunityType.SOO:
            return "ip extcommunity-list", match_type, name, "permit", members
        elif comm_type is CommunityType.LARGE:
            return "ip large-community-list", match_type, name, "permit", members
        else:
            raise NotImplementedError(f"CommunityList type {comm_type} not implemented for arista")

    def _arista_community_prefix(self, community_list: CommunityList) -> str:
        if community_list.type is CommunityType.BASIC:
            return ""
        elif community_list.type is CommunityType.RT:
            if community_list.use_regex:
                return "RT:"
            return "rt "
        elif community_list.type is CommunityType.SOO:
            if community_list.use_regex:
                return "SoO:"
            return "soo "
        elif community_list.type is CommunityType.LARGE:
            return ""
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for arista")

    def run_arista(self, device):
        for community_list_union in self.get_used_united_community_lists(device):
            name = mangle_united_community_list_name([c.name for c in community_list_union])
            for community_list in community_list_union:
                if community_list.use_regex and len(community_list.members) > 1:
                    raise NotImplementedError("Multiple regex is not supported for arista")

                member_prefix = self._arista_community_prefix(community_list)

                if community_list.logic == CommunityLogic.AND:
                    # to get AND logic the communities should be in one sting
                    member_str = " ".join(
                        member_prefix + arista_well_known_community(m)
                        for m in community_list.members
                    )
                    yield self._arista_community_list(
                        name=name,
                        use_regex=community_list.use_regex,
                        comm_type=community_list.type,
                        members=member_str,
                    )
                elif community_list.logic == CommunityLogic.OR:
                    for member in community_list.members:
                        yield self._arista_community_list(
                            name=name,
                            use_regex=community_list.use_regex,
                            comm_type=community_list.type,
                            members=member_prefix + member,
                        )
                else:
                    raise NotImplementedError(f"Community logic {community_list.logic} is not implemented for arista")

    def acl_iosxr(self, _) -> str:
        return r"""
        community-set *
            ~ %global=1
        extcommunity-set *
            ~ %global=1
        """

    def _iosxr_community_list(self, community_list: CommunityList) -> Iterator[Sequence[str]]:
        if community_list.type is CommunityType.BASIC:
            name = "community-set"
        elif community_list.type is CommunityType.RT:
            name = "extcommunity-set rt"
        elif community_list.type is CommunityType.SOO:
            name = "extcommunity-set soo"
        else:
            raise NotImplementedError(f"CommunityList type {community_list.type} not implemented for Cisco IOS XR")

        with self.block(name, community_list.name):
            for n, community in enumerate(community_list.members):
                if n + 1 < len(community_list.members):
                    comma = ","
                else:
                    comma = ""
                if community_list.use_regex:
                    yield "ios-regex", f"'{community}'"
                else:
                    yield f"{community}{comma}",

    def run_iosxr(self, device):
        for community_list in self.get_used_community_lists(device):
            yield from self._iosxr_community_list(community_list)

    def acl_juniper(self, _) -> str:
        return r"""
        policy-options  %cant_delete
            community ~
        """

    def _juniper_community_list(self, name: str, community_lists: list[CommunityList]) -> Iterator[Sequence[str]]:
        members: list[str] = []
        logic: set[CommunityLogic] = set()
        for community_list in community_lists:
            prefix: str
            if community_list.type is CommunityType.BASIC:
                prefix = ""
            elif community_list.type is CommunityType.RT:
                prefix = "target:"
            elif community_list.type is CommunityType.SOO:
                prefix = "origin:"
            elif community_list.type is CommunityType.LARGE:
                prefix = "large:"
            else:
                raise NotImplementedError(f"CommunityList {name}: type {community_list.type} not implemented for Juniper")

            logic.add(community_list.logic)
            for community in community_list.members:
                members.append(prefix + community)

        if len(members) > 1 and logic != {CommunityLogic.AND}:
            raise NotImplementedError(f"CommunityList {name}: only AND logic between members is implemeted for Juniper")

        definition = ["community", name, "members"]
        with self.block("policy-options"):
            if len(members) == 1:
                yield *definition, *members
            if len(members) > 1:
                yield *definition, "[", *members, "]"

    def run_juniper(self, device):
        # Juniper allows different community types
        # so we write generator in a generic way to reflect that.
        #
        # But get_used_community_lists DOES NOT allow multiple names
        # This is in part because juniper does not have a type-aware match
        # It would mean that there is no way to describe a following config via rpl.py:
        #
        #    CommunityList("COMM_LIST", ["65000:4000"], CommunityType.BASIC),
        #    CommunityList("COMM_LIST", ["65000:4000"], CommunityType.RT),
        #
        #    # match only route-target but not basic one
        #    with route(R.extcommunity_rt.has("COMM_LIST")) as rule:
        #       ...
        used = self.get_used_community_lists(device)
        by_name: Mapping[str, list[CommunityList]] = defaultdict(list)
        for community_list in used:
            by_name[community_list.name].append(community_list)

        for name, community_lists in by_name.items():
            yield from self._juniper_community_list(name, community_lists)



from collections.abc import Sequence
from dataclasses import dataclass
from typing import Optional

from .action import Action
from .condition import AndCondition
from .result import ResultType


@dataclass
class RoutingPolicyStatement:
    name: Optional[str]
    number: Optional[int]
    match: AndCondition
    then: Action
    result: ResultType


@dataclass
class RoutingPolicy:
    name: str
    statements: Sequence[RoutingPolicyStatement]



import enum
from operator import itemgetter

from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import CommonFormatter

from .base import AbstractVendor


_SENTINEL = enum.Enum("_SENTINEL", "sentinel")
sentinel = _SENTINEL.sentinel


class GenericVendor(AbstractVendor):
    def match(self) -> list[str]:
        return []

    @property
    def reverse(self) -> str:
        return "-"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("")

    def make_formatter(self, **kwargs) -> CommonFormatter:
        return CommonFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return ""


GENERIC_VENDOR = GenericVendor()


class Registry:
    def __init__(self):
        self.vendors: dict[str, AbstractVendor] = {}
        self._matchers = {}

    def register(self, cls: type[AbstractVendor]) -> type[AbstractVendor]:
        if not cls.NAME:
            raise RuntimeError(f"{cls.__name__} has empty NAME field")
        if cls.NAME in self.vendors:
            raise RuntimeError(f"{cls.__name__} with name {cls.NAME} already registered")
        self.vendors[cls.NAME] = cls()

        return cls

    def __add__(self, other: "Registry"):
        self.vendors = dict(**other.vendors, **self.vendors)

    def __getitem__(self, item) -> AbstractVendor:
        if item in self.vendors:
            return self.vendors[item]
        raise RuntimeError(f"Unknown vendor {item}")

    def match(
        self,
        hw: HardwareView | str,
        default: _SENTINEL | AbstractVendor | None = sentinel
    ) -> AbstractVendor | None:
        if isinstance(hw, str):
            hw = HardwareView(hw, "")

        matched: list[tuple[AbstractVendor, int]] = []
        for name, vendor in self.vendors.items():
            for item in vendor.match():
                if hw.match(item):
                    matched.append((vendor, item.count(".")))

        if matched:
            return next(iter(sorted(matched, key=itemgetter(1), reverse=True)))[0]
        if default is sentinel:
            return GENERIC_VENDOR
        return default

    def get(self, item: str, default: _SENTINEL | AbstractVendor | None = sentinel) -> AbstractVendor | None:
        if item in self:
            return self[item]
        if default is sentinel:
            return GENERIC_VENDOR
        return default

    def __contains__(self, item):
        return item in self.vendors

    def __iter__(self):
        return iter(self.vendors)


registry = Registry()



from dataclasses import dataclass, field
from datetime import datetime, timezone
from ipaddress import ip_interface, IPv6Interface
from typing import List, Optional, Any, Dict, Sequence

from annet.annlib.netdev.views.dump import DumpableView
from annet.annlib.netdev.views.hardware import HardwareView, lag_name
from annet.storage import Storage
from annet.vendors import registry_connector


@dataclass
class Entity(DumpableView):
    id: int
    name: str

    @property
    def _dump__list_key(self):
        return self.name


@dataclass
class Label:
    value: str
    label: str


@dataclass
class IpFamily:
    value: int
    label: str


@dataclass
class DeviceType:
    id: int
    manufacturer: Entity
    model: str


@dataclass
class DeviceIp(DumpableView):
    id: int
    display: str
    address: str
    family: int

    @property
    def _dump__list_key(self):
        return self.address


@dataclass
class Prefix(DumpableView):
    id: int
    prefix: str
    site: Optional[Entity]
    vrf: Optional[Entity]
    tenant: Optional[Entity]
    vlan: Optional[Entity]
    role: Optional[Entity]
    status: Label
    is_pool: bool
    custom_fields: dict[str, Any]
    created: datetime
    last_updated: datetime
    description: Optional[str] = ""

    @property
    def _dump__list_key(self):
        return self.prefix


@dataclass
class IpAddress(DumpableView):
    id: int
    assigned_object_id: int
    display: str
    family: IpFamily
    address: str
    status: Label
    tags: List[Entity]
    created: datetime
    last_updated: datetime
    prefix: Optional[Prefix] = None
    vrf: Optional[Entity] = None

    @property
    def _dump__list_key(self):
        return self.address


@dataclass
class InterfaceConnectedEndpoint(Entity):
    device: Entity


@dataclass
class InterfaceType:
    value: str
    label: str


@dataclass
class InterfaceMode:
    value: str
    label: str


@dataclass
class InterfaceVlan(Entity):
    vid: int


@dataclass
class Interface(Entity):
    device: Entity
    enabled: bool
    description: str
    type: InterfaceType
    connected_endpoints: Optional[list[InterfaceConnectedEndpoint]]
    mode: Optional[InterfaceMode]
    untagged_vlan: Optional[InterfaceVlan]
    tagged_vlans: Optional[List[InterfaceVlan]]
    display: str = ""
    ip_addresses: List[IpAddress] = field(default_factory=list)
    vrf: Optional[Entity] = None
    mtu: int | None = None
    lag: Entity | None = None
    lag_min_links: int | None = None

    def add_addr(self, address_mask: str, vrf: str | None) -> None:
        addr = ip_interface(address_mask)
        if vrf is None:
            vrf_obj = None
        else:
            vrf_obj = Entity(id=0, name=vrf)

        if isinstance(addr, IPv6Interface):
            family = IpFamily(value=6, label="IPv6")
        else:
            family = IpFamily(value=4, label="IPv4")

        for existing_addr in self.ip_addresses:
            if existing_addr.address == address_mask and (
                (existing_addr.vrf is None and vrf is None) or
                (existing_addr.vrf is not None and existing_addr.vrf.name == vrf)
            ):
                return
        self.ip_addresses.append(IpAddress(
            id=0,
            display=address_mask,
            address=address_mask,
            vrf=vrf_obj,
            prefix=None,
            family=family,
            created=datetime.now(timezone.utc),
            last_updated=datetime.now(timezone.utc),
            tags=[],
            status=Label(value="active", label="Active"),
            assigned_object_id=self.id,
        ))


@dataclass
class NetboxDevice(Entity):
    url: str
    storage: Storage

    display: str
    device_type: DeviceType
    device_role: Entity
    tenant: Optional[Entity]
    platform: Optional[Entity]
    serial: str
    asset_tag: Optional[str]
    site: Entity
    rack: Optional[Entity]
    position: Optional[float]
    face: Optional[Label]
    status: Label
    primary_ip: Optional[DeviceIp]
    primary_ip4: Optional[DeviceIp]
    primary_ip6: Optional[DeviceIp]
    tags: List[Entity]
    custom_fields: Dict[str, Any]
    created: datetime
    last_updated: datetime

    fqdn: str
    hostname: str
    hw: HardwareView
    breed: str

    interfaces: List[Interface]

    @property
    def neighbors(self) -> List["Entity"]:
        return [
            endpoint.device
            for iface in self.interfaces
            if iface.connected_endpoints
            for endpoint in iface.connected_endpoints
            if endpoint.device
        ]

    # compat
    @property
    def neighbours_fqdns(self) -> list[str]:
        return [dev.name for dev in self.neighbors]

    @property
    def neighbours_ids(self):
        return [dev.id for dev in self.neighbors]

    def __hash__(self):
        return hash((self.id, type(self)))

    def __eq__(self, other):
        return type(self) is type(other) and self.url == other.url

    def is_pc(self) -> bool:
        custom_breed_pc = ("Mellanox", "NVIDIA", "Moxa", "Nebius")
        return self.device_type.manufacturer.name in custom_breed_pc or self.breed == "pc"

    def _make_interface(self, name: str, type: InterfaceType) -> Interface:
        return Interface(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )

    def _lag_name(self, lag: int) -> str:
        return lag_name(self.hw, lag)

    def make_lag(self, lag: int, ports: Sequence[str], lag_min_links: int | None) -> Interface:
        new_name = self._lag_name(lag)
        for target_interface in self.interfaces:
            if target_interface.name == new_name:
                return target_interface
        lag_interface = self._make_interface(
            name=new_name,
            type=InterfaceType(value="lag", label="Link Aggregation Group (LAG)"),
        )
        lag_interface.lag_min_links = lag_min_links
        for interface in self.interfaces:
            if interface.name in ports:
                interface.lag = lag_interface
        self.interfaces.append(lag_interface)
        return lag_interface

    def _svi_name(self, svi: int) -> str:
        return registry_connector.get().match(self.hw).svi_name(svi)

    def add_svi(self, svi: int) -> Interface:
        name = self._svi_name(svi)
        for interface in self.interfaces:
            if interface.name == name:
                return interface
        interface = self._make_interface(
            name=name,
            type=InterfaceType("virtual", "Virtual")
        )
        self.interfaces.append(interface)
        return interface

    def _subif_name(self, interface: str, subif: int) -> str:
        return f"{interface}.{subif}"

    def add_subif(self, interface: str, subif: int) -> Interface:
        name = self._subif_name(interface, subif)
        for target_port in self.interfaces:
            if target_port.name == name:
                return target_port
        target_port = self._make_interface(
            name=name,
            type=InterfaceType("virtual", "Virtual")
        )
        self.interfaces.append(target_port)
        return target_port

    def find_interface(self, name: str) -> Optional[Interface]:
        for iface in self.interfaces:
            if iface.name == name:
                return iface
        return None



import copy
import re
from collections import namedtuple

from annet.annlib.types import Op
from contextlog import get_logger

from annet.rulebook import common


class VRPVersion(namedtuple("VRPVersionBase", ["V", "R", "C", "SPC"])):
    ANY = object()
    ATTR_NAMES = ["V", "R", "C", "SPC"]

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False

        for attr_name in self.ATTR_NAMES:
            self_attr = getattr(self, attr_name)
            if self_attr is self.ANY:
                continue
            other_attr = getattr(other, attr_name)
            if other_attr is self.ANY:
                continue

            if self_attr != other_attr:
                return False

        return True

    def __ne__(self, other):
        return not self == other


def parse_version(version: str) -> VRPVersion:
    # CP - Cold Patch
    # HP - Hot Patch
    if not version:
        # FIXME: возможно, если в RT нет данных, надо спрашивать у самого устройства?
        version = "VRP V200R002C50SPC800"
        get_logger().warning("SW version not set, falling back to %r", version)
    res = re.match(r"(?:VRP )?V(?P<v>\d+)R(?P<r>\d+)C(?P<c>\d+)(SPC(?P<spc>\d+))?(?P<opt>T)?", version)
    assert res is not None, f"can't parse version '{version}'"
    m = res.groupdict()  # pylint: disable=invalid-name
    return VRPVersion(int(m["v"]), int(m["r"]), int(m["c"] or 0), int(m["spc"] or 0))


# =====
def rp_node(rule, key, diff, **_):
    # route-policy NAME ACTION node NUM
    (rp_name, node_id) = key
    if diff[Op.REMOVED]:
        if diff[Op.ADDED]:
            sub_diff = {Op.AFFECTED: [], Op.ADDED: [], Op.REMOVED: [], Op.MOVED: [], Op.UNCHANGED: []}
            sub_diff[Op.AFFECTED] = diff[Op.REMOVED]
            yield from common.default(rule, key, sub_diff)
        else:
            yield (False, "undo route-policy %s node %s" % (rp_name, node_id), None)

    if diff[Op.AFFECTED] or diff[Op.ADDED]:
        yield from common.default(rule, key, diff)


def undo_redo(rule, key, diff, **_):
    yield from common.undo_redo(rule, key, diff, **_)


def prefix_list(rule, key, diff, **kwargs):
    # для того чтобы опредилить полностью ли изменяется
    # префикс лист в рулбуке huawei.rul описан ключ (family, name)
    # однако с точки зрения команды каждый индекс - отдельная команда
    # поэтому мы группируем их по индексу тут и передаем в common
    diff_by_index = {}
    for op, rows in diff.items():
        for row in rows:
            # ожидаемый формат команды префикс-листа
            # ip ip-prefix PRFX_CT_LU_ALLOWED_ROUTES index 15 ..
            # ip ipv6-prefix PFXS_SPECIALv6 index 20 ..
            _ip, _family, _name, _index, index, *_ = row["row"].split()
            if index not in diff_by_index:
                sub_diff = {op: [] for op in diff.keys()}
                diff_by_index[index] = sub_diff
            diff_by_index[index][op].append(row)

    family, name = key
    if family not in {"ip", "ipv6"}:
        raise NotImplementedError("Unknown family '%s'" % family)
    if diff[Op.ADDED] or diff[Op.REMOVED] or diff[Op.MOVED]:
        # поскольку исходно у нас в ключе правила нет индекса
        # нужно добавить его туда иначе undo-правило будет без оного
        indexed_rule = copy.deepcopy(rule)
        indexed_rule["reverse"] = "undo ip {}-prefix {} index {}"

        # stub_index референсится в рулбуке huawei.order чтобы обеспечить
        # добавление/удаление стаба в первую/последнюю очередь
        stub, stub_index = "", 99999999

        # если мы только добавляем новые команды (например создаем) в префик-лист
        # либо удаляем/двигаем но при этом у нас есть не изменяемые части
        # хуавей не будет считать лист удаляемым и стаб-правило не нужно
        if (diff[Op.REMOVED] or diff[Op.MOVED]) and not diff[Op.UNCHANGED]:
            stub = "deny 0.0.0.0 32" if family == "ip" else "deny :: 128"
        if stub:
            yield (True, f"ip {family}-prefix {name} index {stub_index} {stub}", None)
        for index, sub_diff in diff_by_index.items():
            yield from common.undo_redo(indexed_rule, (family, name, index), sub_diff, **kwargs)
        if stub:
            yield (False, f"undo ip {family}-prefix {name} index {stub_index}", None)


def static(rule, key, diff, **_):
    """
    Для отката статического маршрута фактически необходимо передавать почти все аргументы,
    кроме различных track ...
    При этом, аргументов может быть разное количество - опциональный VRF, опциональный интерфейс.
    Поэтому мы не парсим саму команду, а только удаляем ненужные аргументы.
    """
    if diff[Op.REMOVED]:
        param = key[0]
        idx = param.find(" track")
        if idx > 0:
            key = (param[0:idx],)
        idx = param.find(" description")
        if idx > 0:
            key = (param[0:idx],)
        idx = param.find(" preference")
        if idx > 0:
            key = (param[0:idx],)
    yield from common.default(rule, key, diff)


def undo_trust(rule, key, diff, hw, **_):
    """на CE свитчах команда undo trust; на S undo trust *"""
    if diff[Op.REMOVED]:
        if hw.Quidway and not hw.S6700:
            yield False, "undo trust %s" % key, None
        else:
            yield False, "undo trust", None
    else:
        yield from common.default(rule, key, diff)


def port_queue(rule, key, diff, **_):
    """
    Для отката конфигурации port-queue на интерфейсе требуется только частичное указание параметров.
    Пример отключения/включения:
    interface 100GE0/1/33
        undo port-queue af3 wfq outbound
        port-queue af3 wfq weight 30 port-wred WRED outbound

    По сути на убрать все параметры между 'wfq' и 'outbound'
    NOC-19414
    """
    if diff[Op.REMOVED]:
        param = key[0]
        idx = param.find("weight")
        if idx > 0:
            key = (param[0:idx] + "outbound",)
    yield from common.default(rule, key, diff)


def netstream_undo(rule, key, diff, **_):
    if diff[Op.REMOVED]:
        # The only part we need is the last keyword: inbound or outbound
        # Unfortunately, key is a tuple so we cast it to a list and back
        key = list(key)
        key[1] = key[1].split(" ")[-1]
        key = tuple(key)
    yield from common.default(rule, key, diff)


def old_snmp_iface_trap_undo(rule, key, diff, hw, **_):
    # хитрая логика для старый хуавеев
    # тут вместо полной команды с undo нужно сгенерить не полную строку
    if diff[Op.REMOVED]:
        if hw.Quidway:
            yield False, "undo mac-address trap notification", None
        else:
            yield False, "undo mac-address trap notification learn", None
    else:
        yield from common.default(rule, key, diff)


def stelnet(rule, key, diff, **_):
    # не заменяем строки stelnet ipv4 server enable и stelnet ipv6 server enable на stelnet server enable
    # чтобы не дергать SSH
    if diff[Op.REMOVED] and diff[Op.ADDED]:
        removed = {x["row"] for x in diff[Op.REMOVED]}
        added = {x["row"] for x in diff[Op.ADDED]}
        if removed == {"stelnet ipv4 server enable", "stelnet ipv6 server enable"} and added == {"stelnet server enable"}:
            return
    yield from common.default(rule, key, diff)


def snmpagent_sysinfo_version(rule, key, diff, hw, **_):
    if hw.Huawei.CE and (diff[Op.ADDED] or diff[Op.REMOVED]):
        assert len(diff[Op.AFFECTED]) == 0, "WTF? Affected not empty: %r" % (diff[Op.AFFECTED])
        versions = set(["v1", "v2c", "v3"])

        result = set()
        for op in [Op.REMOVED, Op.ADDED]:
            for action in diff[op]:
                args = action["row"].split()[3:]
                assert len(args) > 0, "Empty op %r: %r" % (op, action["row"])

                if args[-1] == "disable":
                    args = args[:-1]
                    disable = True
                else:
                    disable = False
                if "all" in args:
                    args = versions
                else:
                    assert len(set(args).difference(versions)) == 0, "Incorrect args: %r" % (args)

                if (op == Op.REMOVED and disable) or (op == Op.ADDED and not disable):
                    result.update(args)
                else:
                    result.difference_update(args)

        if result == versions:
            yield (True, "snmp-agent sys-info version all", None)
        else:
            yield (False, "snmp-agent sys-info version all disable", None)
            yield (True, "snmp-agent sys-info version %s" % (" ".join(result)), None)
    else:
        yield from common.default(rule, key, diff)


def vty_acl_undo(rule, key, diff, **_):
    if diff[Op.REMOVED]:
        chunks = key[0].split()
        result_chunks = ["undo acl"]
        if len(chunks) == 3 and chunks[0] == "ipv6":
            result_chunks.append("ipv6")
        result_chunks.append(chunks[-1])
        yield False, " ".join(result_chunks), None
    else:
        yield from common.default(rule, key, diff)


def port_split(rule, key, diff, **_):
    # pylint: disable=unused-argument
    def _port_split(old, new, old_row, new_row):
        removed = set(old).difference(new)
        added = set(new).difference(old)
        if old and new:
            for ifname in removed:
                yield (False, "undo port split dimension interface " + ifname, None)
            for ifname in added:
                yield (True, "port split dimension interface " + ifname, None)
        elif old and not new:
            yield (False, "undo " + old_row, None)
        elif new and not old:
            yield (True, new_row, None)

    def _row_slot(row):
        res = ""
        for ch in row:
            if ch == "/":
                break
            res = res + ch if ch.isnumeric() else ""
        return int(res) if res else 0

    old_by_slot = {_row_slot(x["row"]): x["row"] for x in diff[Op.REMOVED]}
    new_by_slot = {_row_slot(x["row"]): x["row"] for x in diff[Op.ADDED]}
    for slot in set(old_by_slot.keys()).union(new_by_slot.keys()):
        old_row = old_by_slot[slot] if slot in old_by_slot else ""
        new_row = new_by_slot[slot] if slot in new_by_slot else ""
        old = _expand_portsplit(old_row)
        new = _expand_portsplit(new_row)
        yield from _port_split(old, new, old_row, new_row)
    if old_by_slot or new_by_slot:
        yield (True, "port split refresh", None)


def _expand_portsplit(row):
    expanded = []
    row_parts = row.split()
    for (index, part) in enumerate(row_parts):
        if part == "to":
            iface_base = "/".join(row_parts[index - 1].split("/")[:-1])
            left = int(row_parts[index - 1].split("/")[-1])
            right = int(row_parts[index + 1].split("/")[-1])
            for i in range(left + 1, right):
                expanded.append(iface_base + "/" + str(i))
        else:
            expanded.append(part)
    return expanded


def classifier(rule, key, diff, **_):
    # если type меняется нужно сначала удалить все if-match
    # а после этого пересоздать classifier
    if diff[Op.ADDED] and diff[Op.REMOVED]:
        yield (True, diff[Op.REMOVED][0]["row"], diff[Op.REMOVED][0]["children"])
    yield from common.default(rule, key, diff)


def undo_children(rule, key, diff, **_):
    def removed_count(subdiff):
        ret = 0
        for child in subdiff["children"].values():
            for child_diff in child["items"].values():
                ret += len(child_diff[Op.REMOVED])
        return ret

    def common_default(op, subdiff):
        newdiff = {Op.ADDED: [], Op.REMOVED: [], Op.MOVED: [], Op.AFFECTED: [], Op.UNCHANGED: []}
        newdiff[op] = [subdiff]
        yield from common.default(rule, key, newdiff)

    # Приходится самим говорить undo поскольку мы притворяемся одним блоком
    for subdiff in diff[Op.REMOVED]:
        # Сначала нужно удалить все group-member'ы
        if diff[Op.REMOVED][0]["children"]:
            yield (True, diff[Op.REMOVED][0]["row"], diff[Op.REMOVED][0]["children"])
        yield False, "undo " + subdiff["row"], None
    # Сначала разбираем affected, там внутри могут быть undo
    for subdiff in sorted(diff[Op.AFFECTED], key=removed_count, reverse=True):
        yield from common_default(Op.AFFECTED, subdiff)
    for subdiff in diff[Op.ADDED]:
        yield from common_default(Op.ADDED, subdiff)


def clear_instead_undo(rule, key, diff, **_):
    # Для ряда конфигурационных строк возникает вечный diff, поскольку в конфиге строка либо явно включена,
    # либо явно выключена. Если она не описана в генераторе, т.е. мы полагаемся на дефолт, то используя clear
    # вместо undo мы возвращаем конфиг в дефолтное состояние.
    # NOC-20102 @gslv 11-02-2022
    if diff[Op.REMOVED]:
        if diff[Op.REMOVED][0]["row"].endswith(" disable"):
            cmd = diff[Op.REMOVED][0]["row"].replace(" disable", "")
        yield (True, "clear " + cmd, False)
    else:
        yield from common.default(rule, key, diff)



import itertools
from typing import List

from annet.generators import BaseGenerator
from annet.storage import Storage

from . import lldp
from . import hostname


def get_generators(store: Storage) -> List[BaseGenerator]:
    return list(itertools.chain.from_iterable([
        hostname.get_generators(store),
        lldp.get_generators(store),
    ]))




import re
import tempfile
import os
import sys
import math
import asyncio
import platform
import textwrap
import time
import traceback
from contextlib import contextmanager
from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Optional, Any

from contextlog import get_logger

from annet import text_term_format
from annet.deploy import ProgressBar
from annet.output import TextArgs
try:
    import curses
except ImportError:
    curses = None

uname = platform.uname()[0]
NCURSES_SIZE_T = 2 ** 15 - 1
MIN_CONTENT_HEIGHT = 20


class AskConfirm:
    CUT_WARN_MSG = "WARNING: the text was cut because of curses limits."

    def __init__(self, text: str, text_type="diff", alternative_text: str = "",
                 alternative_text_type: str = "diff", allow_force_yes: bool = False):
        self.text = [text, text_type]
        self.alternative_text = [alternative_text, alternative_text_type]
        self.color_to_curses: Dict[Optional[str], int] = {}
        self.lines: Dict[int, List[TextArgs]] = {}
        self.rows = 0
        self.cols = None
        self.top = 0
        self.left = 0
        self.pad: Optional["curses.window"] = None
        self.screen = None
        self.found_pos: dict[int, list[TextArgs]] = {}
        self.curses_lines = None
        self.debug_prompt = TextArgs("")
        self.page_position = TextArgs("")
        s_force = "/f" if allow_force_yes else ""
        self.prompt = [
            TextArgs("Execute these commands? [Y%s/q] (/ - search, a - patch/cmds)" % s_force, "blue", offset=0),
            self.page_position,
            self.debug_prompt]

    def _parse_text(self):
        txt = self.text[0]
        txt_split = txt.splitlines()
        # curses pad, который тут используется, имеет ограничение на количество линий
        if (len(txt_split) + 1) >= NCURSES_SIZE_T:  # +1 для того чтобы курсор можно было переместить на пустую строку
            del txt_split[NCURSES_SIZE_T - 3:]
            txt_split.insert(0, self.CUT_WARN_MSG)
            txt_split.append(self.CUT_WARN_MSG)
            txt = "\n".join(txt_split)
        self.rows = len(txt_split)
        self.cols = max(len(line) for line in txt_split)
        res = text_term_format.curses_format(txt, self.text[1])
        self.lines = res

    def _update_search_pos(self, pattern: str) -> None:
        self.found_pos = {}
        if not pattern:
            return
        try:
            expr = re.compile(pattern)
        except Exception:
            return None
        lines = self.text[0].splitlines()
        for (line_no, line) in enumerate(lines):
            for match in re.finditer(expr, line):
                if line_no not in self.found_pos:
                    self.found_pos[line_no] = []
                self.found_pos[line_no].append(TextArgs(match.group(0), "highlight", match.start()))

    def _init_colors(self):
        self.color_to_curses = init_colors()

    def _init_pad(self):
        import curses

        with self._store_xy():
            self.pad = curses.newpad(self.rows + 1, self.cols)
            self.pad.keypad(True)  # accept arrow keys
            self._render_to_pad(self.lines)

    def _render_to_pad(self, lines: dict):
        """
        Рендерим данный на pad
        :param lines: словарь проиндексированный по номерам линий
        :return:
        """
        with self._store_xy():
            for line_no, line_data in sorted(lines.items()):
                line_pos_calc = 0
                for line_part in line_data:
                    if line_part.offset is not None:
                        line_pos = line_part.offset
                    else:
                        line_pos = line_pos_calc
                    if line_part.color:
                        self.pad.addstr(line_no, line_pos, line_part.text, self.color_to_curses[line_part.color])
                    else:
                        self.pad.addstr(line_no, line_pos, line_part.text)
                    line_pos_calc += len(line_part.text)

    def _add_prompt(self):
        for prompt_part in self.prompt:
            if not prompt_part:
                continue
            if prompt_part.offset is None:
                offset = 0
            else:
                offset = prompt_part.offset
            self.screen.addstr(self.curses_lines - 1, offset, prompt_part.text, self.color_to_curses[prompt_part.color])

    def _clear_prompt(self):
        with self._store_xy():
            self.screen.move(self.curses_lines - 1, 0)
            self.screen.clrtoeol()

    def show(self):
        self._add_prompt()
        self.screen.refresh()
        size = self.screen.getmaxyx()
        self.pad.refresh(self.top, self.left, 0, 0, size[0] - 2, size[1] - 2)

    @contextmanager
    def _store_xy(self):
        if self.pad is not None:
            current_y, current_x = self.pad.getyx()
            yield current_y, current_x
            max_y, max_x = self.pad.getmaxyx()
            current_y = min(max_y - 1, current_y)
            current_x = min(max_x - 1, current_x)

            self.pad.move(current_y, current_x)
        else:
            yield

    def search_next(self, prev=False):
        to = None
        current_y, current_x = self.pad.getyx()
        if prev:
            for line_index in sorted(self.found_pos, reverse=True):
                for text_args in self.found_pos[line_index]:
                    if line_index > current_y:
                        continue

                    if line_index < current_y or line_index == current_y and text_args.offset < current_x:
                        to = line_index, text_args.offset
                        break
                if to:
                    break
        else:
            for line_index in sorted([i for i in self.found_pos if i >= current_y]):
                for text_args in self.found_pos[line_index]:
                    if line_index > current_y or line_index == current_y and text_args.offset > current_x:
                        to = line_index, text_args.offset
                        break
                if to:
                    break
        if to:
            return to[0] - current_y, to[1] - current_x
        else:
            return 0, 0

    def _search_prompt(self) -> tuple[int, int]:
        import curses

        search_prompt = [TextArgs("Search: ", "green_bold", offset=0)]
        current_prompt = self.prompt
        self.prompt = search_prompt
        with self._store_xy():
            self._clear_prompt()
            self.show()
            curses.echo()
            expr = self.screen.getstr().decode()
            curses.noecho()
            self._update_search_pos(expr)
            self._parse_text()
            self._init_pad()
            # срендерем поверх pad слой с подстветкой
            self._render_to_pad(self.found_pos)
            y_offset, x_offset = self.search_next()
        self.prompt = current_prompt
        return y_offset, x_offset

    def _do_commands(self) -> str:
        import curses

        while True:
            self._clear_prompt()
            try:
                ch = self.pad.getch()
            except KeyboardInterrupt:
                return "n"
            max_y, max_x = self.screen.getmaxyx()
            _, pad_max_x = self.pad.getmaxyx()
            max_y -= 2  # prompt
            y_offset = 0
            x_offset = 0
            margin = 0
            y_delta = 0
            x_delta = 0

            y, x = self.pad.getyx()
            if ch == ord("q"):
                return "exit"
            elif ch in [ord("y"), ord("Y")]:
                return "y"
            elif ch in [ord("f"), ord("F")]:
                return "force-yes"
            elif ch == ord("a"):
                if self.alternative_text:
                    self.text, self.alternative_text = self.alternative_text, self.text
                self.screen.clear()
                self._parse_text()
                self._init_pad()
            elif ch == ord("d"):
                if self.debug_prompt.text == "":
                    self.debug_prompt.text = "init"
                else:
                    self.debug_prompt.text = ""
            elif ch == ord("n"):
                y_offset, x_offset = self.search_next()
                margin = 10
            elif ch == ord("N"):
                y_offset, x_offset = self.search_next(prev=True)
                margin = 10
            elif ch == ord("/"):
                y_offset, x_offset = self._search_prompt()
                margin = 10
            elif ch == curses.KEY_UP:
                y_offset = -1
            elif ch == curses.KEY_PPAGE:
                y_offset = -10
            elif ch == curses.KEY_HOME:
                y_offset = -len(self.lines)
            elif ch == curses.KEY_DOWN:
                y_offset = 1
            elif ch == curses.KEY_NPAGE:
                y_offset = 10
            elif ch == curses.KEY_END:
                y_offset = len(self.lines)
            elif ch == curses.KEY_LEFT:
                x_offset = -1
            elif ch == curses.KEY_RIGHT:
                x_offset = 1

            if y_offset or x_offset:
                y = max(0, y + y_offset)
                y = min(self.rows, y)
                x = max(0, x + x_offset)
                x = min(self.cols, x)

                y_delta = y - (self.top + max_y - margin)
                if y_delta > 0:
                    self.top += y_delta
                elif (y - margin) < self.top:
                    self.top = y

                self.top = min(self.top, len(self.lines) - max_y)

                x_delta = x - (self.left + max_x)
                if x_delta > 0:
                    self.left += x_delta
                elif x < self.left:
                    self.left = x

                x = min(x, pad_max_x - 1)
                self.pad.move(y, x)

            if self.debug_prompt.text != "":
                debug_line = "y=%s x=%s, x_delta=%s y_delta=%s top=%s, max_y=%s max_x=%s lines=%s" % \
                             (y, x, x_delta, y_delta, self.top, max_y, max_x, len(self.lines))
                self.debug_prompt.text = debug_line
                self.debug_prompt.color = "green_bold"
                self.debug_prompt.offset = max_x - len(debug_line) - 1

            if self.debug_prompt.text == "":
                self.page_position.color = "highlight"
                self.page_position.text = "line %s/%s" % (y, len(self.lines))
                self.page_position.offset = max_x - len(self.page_position.text) - 1

            self.show()

    def loop(self) -> str:
        import curses

        res = ""
        old_cursor = None
        try:
            self.screen = curses.initscr()
            self.screen.leaveok(True)
            self.curses_lines = curses.LINES  # pylint: disable=maybe-no-member
            curses.start_color()
            curses.noecho()  # no echo key input
            curses.cbreak()  # input with no-enter keyed
            try:
                old_cursor = curses.curs_set(2)
            except Exception:
                pass
            self._init_colors()
            self._parse_text()
            self._init_pad()
            self.pad.move(0, 0)
            self.show()
            res = self._do_commands()
        except Exception as err:
            get_logger().exception("%s", err)
        finally:
            if old_cursor is not None:
                curses.curs_set(old_cursor)
            curses.nocbreak()
            curses.echo()
            curses.endwin()
        return res


def init_colors():
    import curses

    curses.init_pair(1, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(2, curses.COLOR_CYAN, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_MAGENTA, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(6, curses.COLOR_BLUE, curses.COLOR_WHITE)
    curses.init_pair(7, curses.COLOR_RED, curses.COLOR_WHITE)
    curses.init_pair(8, curses.COLOR_BLACK, curses.COLOR_WHITE)
    curses.init_pair(9, curses.COLOR_CYAN, curses.COLOR_BLUE)
    return {
        "green": curses.color_pair(1),
        "green_bold": curses.color_pair(1) | curses.A_BOLD,
        "cyan": curses.color_pair(2),
        "red": curses.color_pair(3),
        "magenta": curses.color_pair(4),
        "yellow": curses.color_pair(5),
        "blue": curses.color_pair(6),
        "highlight": curses.color_pair(7),
        None: curses.color_pair(8),
        "cyan_blue": curses.color_pair(9),
    }


@dataclass
class Tile:
    win: "curses.window | None"
    content: list[list[TextArgs]]
    title: list[str]
    height: int
    width: int
    need_draw: bool = True
    total: int = 0
    iteration: int = 0


class UiState(Enum):
    INIT = "INIT"
    OK = "OK"
    WAIT_INPUT = "WAIT_INPUT"
    CLOSED = "CLOSED"


class TailMode(Enum):
    UNIFORM = "UNIFORM"  # display is split into several parts with content and headers
    ONE_CONTENT = "ONE_CONTENT"  # only one part displays content, others show headers
    NO_CONTENT = "NO_CONTENT"  # only headers


class ProgressBars(ProgressBar):
    def __init__(self, tiles_params: dict[str, dict[Any, Any]]):
        self.tiles_params = tiles_params
        self.mode: TailMode = TailMode.UNIFORM
        self.screen: "curses.window | None" = None
        self.tiles: dict[str, Tile] = {}
        self.offset = [0, 0]
        self.terminal_refresher_coro = None
        self.color_to_curses: dict[Optional[str], int] = {}
        self.state: UiState = UiState.INIT
        self.active_tile: int = len(tiles_params) - 1  # tiles with content have numbers from 1

        # context
        self.enter_ok = False
        self._new_stderr = None
        self._saved_stderr_fd = None
        self.progress_length = 10

    def __enter__(self):
        self.enter_ok = False
        sys.stderr.flush()
        self._new_stderr = tempfile.TemporaryFile()
        self._saved_stderr_fd = os.dup(sys.stderr.fileno())
        os.dup2(self._new_stderr.fileno(), sys.stderr.fileno())
        self.enter_ok = True
        try:
            self.init()
        except Exception:
            self.__exit__(*sys.exc_info())
            raise
        return self

    def __exit__(self, exc_type, exc_value, trace):
        try:
            self.stop_curses()
        except Exception:
            pass
        if self.enter_ok:
            sys.stderr.flush()
            self._new_stderr.seek(0)
            stderr_data = self._new_stderr.read().decode()
            os.dup2(self._saved_stderr_fd, sys.stderr.fileno())
            sys.stderr.write(stderr_data)

    def evaluate_mode(self, scree_size):
        height = scree_size[0] // len(self.tiles_params)
        if height > MIN_CONTENT_HEIGHT:
            return TailMode.UNIFORM
        if scree_size[0] - len(self.tiles_params) > MIN_CONTENT_HEIGHT:
            return TailMode.ONE_CONTENT
        if scree_size[0] // len(self.tiles_params) > 1:
            return TailMode.NO_CONTENT
        else:
            return TailMode.NO_CONTENT

    def make_tiles(self):
        import curses

        i = 0
        scree_size = self.screen.getmaxyx()
        scree_size = scree_size[0] - 1, scree_size[1]
        scree_offset = self.screen.getbegyx()
        mode = self.evaluate_mode(scree_size)
        tiles_count = len(self.tiles_params)
        max_height = scree_size[0]
        width = scree_size[1]
        begin_y = 0
        begin_x = 0
        tile_no = 0
        status_bar_win = curses.newwin(1, width, scree_size[0], 0)
        self.tiles["status:"] = Tile(
            win=status_bar_win, content=[], title=[""],
            height=1, width=width, need_draw=True,
        )
        max_tile_name_len = max(len(tile_name) for tile_name in self.tiles_params)

        for tile_name in self.tiles_params:
            win = None
            height = 0
            tile_no += 1

            if mode is TailMode.UNIFORM:
                height = int(scree_size[0] // len(self.tiles_params))  # TODO:остаток от деления прибавить к последнему
                win = curses.newwin(height, width, begin_y, begin_x)
            elif mode is TailMode.ONE_CONTENT:
                if i == self.active_tile:
                    height = scree_size[0] - tiles_count + 1
                else:
                    height = 1
                win = curses.newwin(height, width, begin_y, begin_x)
            elif mode is TailMode.NO_CONTENT:
                height = 1

                if tile_no < max_height:
                    begin_y = scree_offset[0] + begin_y
                    get_logger().debug("newwin height=%s, width=%s, begin_y=%s, begin_x=%s", height, width, begin_y,
                                       begin_x)
                    win = curses.newwin(height, width, begin_y, begin_x)
                if tile_no == max_height:
                    left = len(self.tiles_params) - max_height + 1
                    self.tiles["dumb"] = Tile(
                        win=curses.newwin(height, width, begin_y, begin_x),
                        content=[],
                        title=["... and %s more" % left],
                        height=height, width=width, need_draw=True,
                    )

            self.tiles[tile_name] = Tile(
                win=win,
                content=[],
                title=[("{:<%s}" % (max_tile_name_len)).format(tile_name)],
                height=height, width=width, need_draw=True,
            )
            i += 1
            begin_y += height

    def _next_active_tile(self):
        self._set_active_tile((self.active_tile + 1) % len(self.tiles_params))

    def _prev_active_tile(self):
        self._set_active_tile((self.active_tile - 1) % len(self.tiles_params))

    def _set_active_tile(self, active_tile: int) -> None:
        self.active_tile = active_tile
        if self.mode is TailMode.ONE_CONTENT:
            return

        scree_size = self.screen.getmaxyx()
        scree_offset = self.screen.getbegyx()
        tiles_count = len(self.tiles_params)
        width = scree_size[1]
        begin_y = scree_offset[0]
        for n, dev in enumerate(self.tiles_params):
            tile = self.tiles[dev]
            if n == self.active_tile:
                height = scree_size[0] - tiles_count
            else:
                height = 1
            tile.height = height
            tile.win.resize(height, width)
            tile.win.mvwin(begin_y, scree_offset[1])
            begin_y += tile.height
            tile.need_draw = True
        self.refresh_all()

    def set_status(self):
        total = 0
        iteration = 0
        done_percent = 0
        for tile_name, tile in self.tiles.items():
            if tile_name == "status:":
                continue
            total += tile.total
            iteration += tile.iteration
        if total > 0 and iteration > 0:
            done_percent = float(iteration) / total * 100

        if done_percent == 100:
            msg = "All is done. Press q for exit."
        else:
            msg = "Deploying... %3.0f%%" % math.floor(done_percent)
        self.set_title("status:", msg)

    def start_terminal_refresher(self, max_refresh_rate=200):
        if self.terminal_refresher_coro:
            return
        self.terminal_refresher_coro = asyncio.ensure_future(self._terminal_refresher(max_refresh_rate))

    def stop_terminal_refresher(self):
        if not self.terminal_refresher_coro:
            return
        self.terminal_refresher_coro.cancel()
        self.terminal_refresher_coro = None
        self.refresh_all()

    async def _terminal_refresher(self, max_refresh_rate: int):
        sleep = 1 / max_refresh_rate
        try:
            while True:
                await asyncio.sleep(sleep)
                self.refresh_all()
        except asyncio.CancelledError:
            return

    def init(self):
        import curses

        self.screen = curses.initscr()
        self.screen.nodelay(True)  # getch() will be non-blocking
        curses.noecho()
        curses.cbreak()
        scree_size = self.screen.getmaxyx()
        get_logger().debug("orig scree_size y=%s, x=%s offset=%s %s", scree_size[0], scree_size[1], self.offset[0],
                           self.offset[1])

        if self.offset[0] == 0:
            new_y = 0
            nlines = scree_size[0]
        elif self.offset[0] < 0:
            nlines = abs(self.offset[0])
            new_y = scree_size[0] - nlines
        else:
            nlines = self.offset[0]
            new_y = scree_size[0]

        if self.offset[1] == 0:
            new_x = 0
            ncols = scree_size[1]
        elif self.offset[1] < 0:
            ncols = abs(self.offset[1])
            new_x = scree_size[1] - ncols
        else:
            ncols = self.offset[1]
            new_x = scree_size[1]

        get_logger().debug("nlines=%s, ncols=%s, new_y=%s, new_x=%s", nlines, ncols, new_y, new_x)

        self.screen.mvwin(new_y, new_x)

        curses.curs_set(0)
        curses.start_color()
        self.color_to_curses = init_colors()
        self.state = UiState.OK
        self.make_tiles()
        self.progress_length = scree_size[1] // 3

    def stop_curses(self):
        import curses

        curses.curs_set(1)
        curses.nocbreak()
        curses.echo()
        curses.endwin()
        self.state = UiState.CLOSED

    def draw_content(self, tile_name):
        tile = self.tiles[tile_name]
        win = tile.win
        size = win.getmaxyx()
        margin = 1
        if (size[0] - 2 * margin) <= 0:
            return
        draw_lines_in_win(tile.content, win, color_to_curses=self.color_to_curses, margin=margin)

    def draw_title(self, tile_name):
        tile = self.tiles[tile_name]
        title = tile.title
        win = tile.win
        if not isinstance(title, (tuple, list)):
            title = [title]
        draw_lines_in_win([title], win, color_to_curses=self.color_to_curses, x_margin=1)

    def refresh(self, tile_name: str, noutrefresh: bool = False):
        # see noutrefresh in curses doc
        tile = self.tiles[tile_name]
        win = tile.win
        if not tile.need_draw or win is None:
            return
        win.clear()
        if tile.height > 1:
            win.border()
        self.draw_title(tile_name)
        self.draw_content(tile_name)
        if noutrefresh:
            win.noutrefresh()
        else:
            win.refresh()
        tile.need_draw = False

    def refresh_all(self):
        if self.state is UiState.CLOSED:
            return
        if self.state is UiState.OK:
            ch_list = self.get_pressed_keys()
            if "\t" in ch_list:  # Tab
                self._next_active_tile()
            if "\x1b[Z" in ch_list:  # shift-Tab
                self._prev_active_tile()
        self.screen.refresh()
        self.set_status()
        tile_name = None
        for tile_name in self.tiles:
            self.refresh(tile_name, True)
        if tile_name is not None:
            self.refresh(tile_name, False)

    def set_title(self, tile_name, title):
        tile = self.tiles[tile_name]
        # в 0 элементе хранится выровненный хостнейм
        title0 = tile.title[0]
        new_title = [title0, title]
        if new_title == tile.title:
            return
        tile.title = new_title
        tile.need_draw = True
        if not self.terminal_refresher_coro:
            self.refresh(tile_name)

    def set_content(self, tile_name: str, content: str):
        tile = self.tiles[tile_name]
        new_content = list(text_term_format.curses_format(content, "switch_out").values())
        if new_content == tile.content:
            return
        tile.need_draw = True
        tile.content = new_content
        if not self.terminal_refresher_coro:
            self.refresh(tile_name)

    def add_content(self, tile_name: str, content: str):
        tile = self.tiles[tile_name]
        tile.need_draw = True
        tile.content.extend(text_term_format.curses_format(content, "switch_out").values())
        if not self.terminal_refresher_coro:
            self.refresh(tile_name)

    def reset_content(self, tile_name: str):
        tile = self.tiles[tile_name]
        tile.need_draw = True
        tile.content = []
        if not self.terminal_refresher_coro:
            self.refresh(tile_name)

    def set_progress(self,
                     tile_name: str,
                     iteration: int,
                     total: int,
                     prefix="",
                     suffix="",
                     fill="█",
                     error=False,
                     ):
        """
        Call in a loop to create terminal progress bar
        @params:
            iteration   - Required  : current iteration (Int)
            total       - Required  : total iterations (Int)
            prefix      - Optional  : prefix string (Str)
            suffix      - Optional  : suffix string (Str)
            length      - Optional  : character length of bar (Int)
            fill        - Optional  : bar fill character (Str)
        """
        if uname == "FreeBSD":
            fill = "#"
        percent = "{0:.1f}".format(100 * (iteration / float(total)))
        filled_length = int(self.progress_length * iteration // total)
        bar = fill * filled_length + "-" * (self.progress_length - filled_length)
        res = "%s |%s| %s%% %s" % (prefix, bar, percent, suffix)
        tile = self.tiles[tile_name]
        tile.total = total
        tile.iteration = iteration
        if error:
            res = TextArgs(res, "red")
        else:
            res = TextArgs(res, "cyan")
        self.set_title(tile_name, res)

    def set_exception(self, tile_name: str, cmd_exc: str, last_cmd: str, progress_max: int, content: str = ""):
        suffix = "cmd error: %s %s" % (str(cmd_exc).strip().replace("\n", "--"), last_cmd)
        self.set_progress(tile_name, progress_max, progress_max, suffix=suffix, error=True)
        if content:
            self.add_content(tile_name, content)

    def get_pressed_keys(self):
        ch_list = ""
        while True:
            try:
                ch = self.screen.getkey()
                ch_list += ch
            except Exception:
                time.sleep(0.01)
                break
        return ch_list

    async def wait_for_exit(self):
        self.state = UiState.WAIT_INPUT
        while True:
            ch_list = self.get_pressed_keys()
            if ch_list:
                get_logger().debug("read ch %s", repr(ch_list))
                if "\t" in ch_list:  # Tab
                    self._next_active_tile()
                if "\x1b[Z" in ch_list:  # shift-Tab
                    self._prev_active_tile()
                if "q" in ch_list:
                    return
            else:
                await asyncio.sleep(0.001)


def draw_lines_in_win(
        lines: list[list[TextArgs]],
        win: "curses.window | None",
        color_to_curses: dict[Optional[str], int],
        margin: int = 0,
        x_margin: int = 0,
        y_margin: int = 0,
) -> None:
    if win is None:
        return
    max_y, max_x = win.getmaxyx()
    max_y -= 2 * (margin or y_margin)
    max_x -= 2 * (margin or x_margin)
    lines_count = len(lines)
    if lines_count > max_y:
        start_line = lines_count - max_y
    else:
        start_line = 0
    for line_no, line_data in enumerate(lines[start_line:]):
        line_pos_calc = 0
        for line_part in line_data:
            if not isinstance(line_part, TextArgs):
                line_part = TextArgs(line_part)
            if line_part.offset is not None:
                line_pos = line_part.offset
            else:
                line_pos = line_pos_calc
            y = (margin or y_margin) + line_no
            x = (margin or x_margin) + line_pos
            max_line_len = max_x - x + 1
            text = line_part.text[0:max_line_len]
            try:
                if line_part.color:
                    win.addstr(y, x, text, color_to_curses[line_part.color])
                else:
                    win.addstr(y, x, text)
            except Exception as exp:
                get_logger().error("y=%s, x=%s, text=%s %s", y, x, text, exp)
            line_pos_calc += len(text)



__all__ = [
    "AsPathFilter",
    "AsPathFilterGenerator",
    "CommunityList",
    "CommunityType",
    "CommunityLogic",
    "CommunityListGenerator",
    "CumulusPolicyGenerator",
    "RoutingPolicyGenerator",
    "RDFilterFilterGenerator",
    "RDFilter",
    "IpPrefixList",
    "IpPrefixListMember",
    "PrefixListFilterGenerator",
    "get_policies",
    "ip_prefix_list",
]

from .aspath import AsPathFilterGenerator
from .community import CommunityListGenerator
from .cumulus_frr import CumulusPolicyGenerator
from .entities import CommunityList, AsPathFilter, CommunityType, CommunityLogic, RDFilter, IpPrefixList, IpPrefixListMember, ip_prefix_list
from .execute import get_policies
from .policy import RoutingPolicyGenerator
from .prefix_lists import PrefixListFilterGenerator
from .rd import RDFilterFilterGenerator



from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import B4comFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class B4ComVendor(AbstractVendor):
    NAME = "b4com"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        if hw.B4com.CS2148P:
            before.add_cmd(Command("conf t"))
            after.add_cmd(Command("end"))
            if do_finalize:
                after.add_cmd(Command("write", timeout=40))
        else:
            before.add_cmd(Command("conf t"))
            if do_commit:
                after.add_cmd(Command("commit"))
                after.add_cmd(Command("end"))
            if do_finalize:
                after.add_cmd(Command("write", timeout=40))

        return before, after

    def match(self) -> list[str]:
        return ["B4com"]

    @property
    def reverse(self) -> str:
        return "no"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("B4com")

    def make_formatter(self, **kwargs) -> B4comFormatter:
        return B4comFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return "exit"

    def svi_name(self, num: int) -> str:
        return f"vlan1.{num}"



from dataclasses import dataclass
from datetime import datetime, timezone
from typing import Optional
from annet.adapters.netbox.common.models import InterfaceType, IpFamily, Label, \
    Prefix, Entity, Interface, IpAddress
from annet.adapters.netbox.v41.models import InterfaceV41, IpAddressV41, \
    NetboxDeviceV41, FHRPGroupAssignmentV41
import annetbox.v42.models


@dataclass
class PrefixV42(Prefix):
    scope: Optional[Entity] = None
    scope_type: str | None = None

    @property
    def site(self) -> Optional[Entity]:
        if self.scope_type == "dcim.site":
            return self.scope
        return None


@dataclass
class IpAddressV42(IpAddress[PrefixV42]):
    role: Optional[Label] = None
    prefix: Optional[PrefixV42] = None


@dataclass
class InterfaceV42(Interface[IpAddressV42, FHRPGroupAssignmentV41]):
    def _add_new_addr(self, address_mask: str, vrf: Entity | None, family: IpFamily) -> None:
        self.ip_addresses.append(IpAddressV42(
            id=0,
            display=address_mask,
            address=address_mask,
            vrf=vrf,
            prefix=None,
            family=family,
            created=datetime.now(timezone.utc),
            last_updated=datetime.now(timezone.utc),
            tags=[],
            status=Label(value="active", label="Active"),
            assigned_object_id=self.id,
        ))


@dataclass
class NetboxDeviceV42(NetboxDeviceV41):
    def __hash__(self):
        return hash((self.id, type(self)))

    def _make_interface(self, name: str, type: InterfaceType) -> InterfaceV42:
        return InterfaceV42(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )


# should be unified in case of netbox update
Vrf = annetbox.v42.models.Vrf
Vlan = annetbox.v42.models.Vlan



import re

from annet.rulebook import common


# Добавление возможности удаления агрегатов, лупбэков, SVI и сабинтерфейсов
def permanent(rule, key, diff, **kwargs):  # pylint: disable=redefined-outer-name
    ifname = key[0]
    # Match group examples:
    # Group 01: Port-Channel10, Loopback1, Vlan800
    # Group 02: Ethernet2/1.20, Port-Channel10.200
    if re.match(r"((?:Port-Channel|Loopback|Vlan)\d+$)|((?:Ethernet|Port-Channel)[\d/]+\.\d+$)", ifname):
        # Эти интерфейсы можно удалять
        yield from common.default(rule, key, diff, **kwargs)
    else:
        yield from common.permanent(rule, key, diff, **kwargs)



from dataclasses import dataclass
from logging import getLogger
from typing import Annotated, Callable, Optional, Union

from annet.bgp_models import Peer, GlobalOptions, BgpConfig
from annet.storage import Device, Storage
from .basemodel import merge, BaseMeshModel, Merge, UseLast, MergeForbiddenError
from .device_models import GlobalOptionsDTO
from .models_converter import to_bgp_global_options, to_bgp_peer, InterfaceChanges, to_interface_changes
from .peer_models import DirectPeerDTO, IndirectPeerDTO, VirtualLocalDTO, VirtualPeerDTO
from .registry import (
    DirectPeer,
    GlobalOptions as MeshGlobalOptions,
    IndirectPeer,
    MatchedDirectPair,
    MeshRulesRegistry,
    MeshSession,
    VirtualLocal,
    VirtualPeer,
)

logger = getLogger(__name__)


@dataclass(frozen=True)
class TargetInterface:
    subif: int | None = None
    svi: int | None = None
    lag: int | None = None
    port: str | None = None


@dataclass(frozen=True)
class PeerKey:
    fqdn: str
    addr: str
    vrf: str
    interface: TargetInterface


def target_interface(
        peer: DirectPeerDTO | IndirectPeerDTO | VirtualLocalDTO,
        ports: list[str],
) -> TargetInterface:
    subif = getattr(peer, "subif", None)
    svi = getattr(peer, "svi", None)
    lag = getattr(peer, "lag", None)
    port = None
    if not (svi or lag):
        if len(ports) == 1:
            port = ports[0]
    return TargetInterface(subif, svi, lag, port)


class Pair(BaseMeshModel):
    local: Annotated[Union[DirectPeerDTO, IndirectPeerDTO], Merge()]
    connected: Annotated[Union[DirectPeerDTO, IndirectPeerDTO], Merge()]
    device: Annotated[Device, UseLast()]
    ports: list[str]


class VirtualPair(BaseMeshModel):
    local: Annotated[VirtualLocalDTO, Merge()]
    connected: Annotated[VirtualPeerDTO, Merge()]


class MeshExecutor:
    def __init__(
            self,
            registry: MeshRulesRegistry,
            storage: Storage,
    ):
        self._registry = registry
        self._storage = storage

    def _execute_globals(self, device: Device) -> GlobalOptionsDTO:
        global_opts = GlobalOptionsDTO()
        for rule in self._registry.lookup_global(device.fqdn):
            handler_name = self._handler_name(rule.handler)
            rule_global_opts = MeshGlobalOptions(rule.match, device)
            logger.debug("Running device handler: %s", handler_name)
            rule.handler(rule_global_opts)

            if rule_global_opts.is_empty():
                # nothing was set
                continue

            try:
                global_opts = merge(global_opts, rule_global_opts)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` global options conflicting with "
                    f"previously loaded for device `{device.fqdn}`:\n" + str(e)
                ) from e
        return global_opts

    def _handler_name(self, handler: Callable) -> str:
        try:
            return f"{handler.__module__}.{handler.__qualname__}"
        except AttributeError:
            return str(handler)

    def _execute_direct_pair(
            self,
            device: Device,
            neighbor_device: Device,
            rule: MatchedDirectPair,
            ports: list[tuple[str, str]],
            all_connected_ports: list[tuple[str, str]],
    ) -> Optional[Pair]:
        session = MeshSession()
        handler_name = self._handler_name(rule.handler)
        logger.debug("Running direct handler: %s", handler_name)
        if rule.direct_order:
            peer_device = DirectPeer(rule.match_left, device, [], [])
            peer_neighbor = DirectPeer(rule.match_right, neighbor_device, [], [])
        else:
            peer_neighbor = DirectPeer(rule.match_left, neighbor_device, [], [])
            peer_device = DirectPeer(rule.match_right, device, [], [])

        for local_port, remote_port in ports:
            peer_device.ports.append(local_port)
            peer_neighbor.ports.append(remote_port)
        for local_port, remote_port in all_connected_ports:
            peer_device.all_connected_ports.append(local_port)
            peer_neighbor.all_connected_ports.append(remote_port)

        if rule.direct_order:
            rule.handler(peer_device, peer_neighbor, session)
        else:
            rule.handler(peer_neighbor, peer_device, session)

        if peer_neighbor.is_empty() and peer_device.is_empty() and session.is_empty():
            # nothing was set
            return None

        try:
            neighbor_dto = merge(DirectPeerDTO(), peer_neighbor, session)
        except MergeForbiddenError as e:
            raise ValueError(
                f"Handler `{handler_name}` provided session data conflicting with "
                f"peer data for device `{neighbor_device.fqdn}`:\n" + str(e)
            ) from e
        try:
            device_dto = merge(DirectPeerDTO(), peer_device, session)
        except MergeForbiddenError as e:
            raise ValueError(
                f"Handler `{handler_name}` provided session data conflicting with "
                f"peer data for device `{device.fqdn}`:\n" + str(e)
            ) from e

        return Pair(local=device_dto, connected=neighbor_dto, device=neighbor_device, ports=peer_device.ports)

    def _execute_direct(self, device: Device) -> list[Pair]:
        # we can have multiple rules for the same pair
        # we merge them according to remote fqdn
        neighbor_peers: dict[PeerKey, Pair] = {}
        rules = self._registry.lookup_direct(device.fqdn, device.neighbours_fqdns)
        fqdns = {
            rule.name_right if rule.direct_order else rule.name_left
            for rule in rules
        }
        logger.debug("Loading neighbor devices: %s", fqdns)
        neighbors = {
            d.fqdn: d for d in self._storage.make_devices(list(fqdns))
        }
        for rule in rules:
            handler_name = self._handler_name(rule.handler)
            if rule.direct_order:
                if rule.name_right not in neighbors:
                    raise ValueError(
                        f"Device `{device.fqdn}` has no neighbor `{rule.name_right}` "
                        f"required for `{handler_name}`. {list(neighbors)}",
                    )
                neighbor_device = neighbors[rule.name_right]
            else:
                if rule.name_left not in neighbors:
                    raise ValueError(
                        f"Device `{device.fqdn}` has no neighbor `{rule.name_left}` required for `{handler_name}`",
                    )
                neighbor_device = neighbors[rule.name_left]
            all_connected_ports = [
                (p1.name, p2.name)
                for p1, p2 in self._storage.search_connections(device, neighbor_device)
            ]
            for ports in rule.port_processor(all_connected_ports):
                pair = self._execute_direct_pair(device, neighbor_device, rule, ports, all_connected_ports)
                if pair is None:
                    # nothing was set
                    continue
                addr = getattr(pair.connected, "addr", None)
                if addr is None:
                    raise ValueError(f"Handler `{handler_name}` returned no peer addr")
                peer_key = PeerKey(
                    fqdn=pair.device.fqdn,
                    addr=addr,
                    vrf=getattr(pair.connected, "vrf", ""),
                    interface=target_interface(pair.local, [p[0] for p in ports]),
                )
                try:
                    if peer_key in neighbor_peers:
                        pair = merge(neighbor_peers[peer_key], pair)
                except MergeForbiddenError as e:
                    if rule.direct_order:
                        pair_names = device.fqdn, pair.device.fqdn
                    else:
                        pair_names = pair.device.fqdn, device.fqdn
                    raise ValueError(
                        f"Handler `{handler_name}` provides data conflicting with "
                        f"previously loaded for device pair {pair_names} "
                        f"with addr={peer_key.addr}, vrf{peer_key.vrf}:\n" + str(e)
                    ) from e
                neighbor_peers[peer_key] = pair
        return list(neighbor_peers.values())

    def _execute_virtual(self, device: Device) -> list[VirtualPair]:
        virtual_peers: list[VirtualPair] = []
        for rule in self._registry.lookup_virtual(device.fqdn):
            for order_number in rule.num:
                handler_name = self._handler_name(rule.handler)
                logger.debug("Running virtual handler: %s", handler_name)
                session = MeshSession()
                peer_device = VirtualLocal(rule.match, device)
                peer_virtual = VirtualPeer(num=order_number)

                rule.handler(peer_device, peer_virtual, session)
                if peer_virtual.is_empty() and peer_device.is_empty() and session.is_empty():
                    # nothing was set
                    continue

                try:
                    virtual_dto = merge(VirtualPeerDTO(), peer_virtual, session)
                except MergeForbiddenError as e:
                    raise ValueError(
                        f"Handler `{handler_name}` provided session data conflicting with "
                        f"virtual peer data for device `{device.fqdn}` and num={order_number}:\n" + str(e)
                    ) from e
                try:
                    device_dto = merge(VirtualLocalDTO(), peer_device, session)
                except MergeForbiddenError as e:
                    raise ValueError(
                        f"Handler `{handler_name}` provided session data conflicting with "
                        f"peer data for device `{device.fqdn}`:\n" + str(e)
                    ) from e

                pair = VirtualPair(local=device_dto, connected=virtual_dto)
                virtual_peers.append(pair)
        return virtual_peers

    def _execute_indirect(self, device: Device, all_fqdns: list[str]) -> list[Pair]:
        # we can have multiple rules for the same pair
        # we merge them according to remote fqdn
        connected_peers: dict[PeerKey, Pair] = {}

        rules = self._registry.lookup_indirect(device.fqdn, all_fqdns)
        fqdns = {
            rule.name_right if rule.direct_order else rule.name_left
            for rule in rules
        }
        logger.debug("Loading indirect connected devices: %s", fqdns)
        connected_devices = {
            d.fqdn: d for d in self._storage.make_devices(list(fqdns))
        }
        for rule in rules:
            session = MeshSession()
            handler_name = self._handler_name(rule.handler)
            logger.debug("Running indirect handler: %s", handler_name)
            if rule.direct_order:
                connected_device = connected_devices[rule.name_right]
                peer_device = IndirectPeer(rule.match_left, device)
                peer_connected = IndirectPeer(rule.match_right, connected_device)
                rule.handler(peer_device, peer_connected, session)
            else:
                connected_device = connected_devices[rule.name_left]
                peer_connected = IndirectPeer(rule.match_left, connected_device)
                peer_device = IndirectPeer(rule.match_right, device)
                rule.handler(peer_connected, peer_device, session)

            if peer_connected.is_empty() and peer_device.is_empty() and session.is_empty():
                # nothing was set
                continue

            try:
                connected_dto = merge(IndirectPeerDTO(), peer_connected, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{connected_device.fqdn}`:\n" + str(e)
                ) from e
            try:
                device_dto = merge(IndirectPeerDTO(), peer_device, session)
            except MergeForbiddenError as e:
                raise ValueError(
                    f"Handler `{handler_name}` provided session data conflicting with "
                    f"peer data for device `{device.fqdn}`:\n" + str(e)
                ) from e

            pair = Pair(local=device_dto, connected=connected_dto, device=connected_device)
            addr = getattr(connected_dto, "addr", None)
            if addr is None:
                raise ValueError(f"Handler `{handler_name}` returned no peer addr")
            peer_key = PeerKey(
                fqdn=connected_device.fqdn,
                addr=addr,
                vrf=getattr(connected_dto, "vrf", ""),
                interface=target_interface(pair.local, []),
            )
            try:
                if peer_key in connected_peers:
                    pair = merge(connected_peers[peer_key], pair)
            except MergeForbiddenError as e:
                if rule.direct_order:
                    pair_names = device.fqdn, connected_device.fqdn
                else:
                    pair_names = connected_device.fqdn, device.fqdn
                raise ValueError(
                    f"Handler `{handler_name}` provides data conflicting with "
                    f"previously loaded for device pair {pair_names} "
                    f"with addr={peer_key.addr}, vrf{peer_key.vrf}:\n" + str(e)
                ) from e
            connected_peers[peer_key] = pair

        return list(connected_peers.values())  # FIXME

    def _to_bgp_peer(self, pair: Pair, interface: Optional[str]) -> Peer:
        return to_bgp_peer(pair.local, pair.connected, pair.device.hostname, interface)

    def _virtual_to_bgp_peer(self, pair: VirtualPair, interface: Optional[str]) -> Peer:
        return to_bgp_peer(pair.local, pair.connected, "", interface)

    def _to_bgp_global(self, global_options: GlobalOptionsDTO) -> GlobalOptions:
        return to_bgp_global_options(global_options)

    def _apply_direct_interface_changes(
            self, device: Device, neighbor: Device, ports: list[str], changes: InterfaceChanges,
    ) -> str:
        # filter ports according to processed in pair
        port_pairs = [
            p
            for p in self._storage.search_connections(device, neighbor)
            if p[0].name in ports
        ]
        if len(port_pairs) > 1:
            if changes.lag is changes.svi is None:
                raise ValueError(
                    f"Multiple connections found between {device.fqdn} and {neighbor.fqdn}."
                    "Specify LAG or SVI"
                )
        if changes.lag is not None:
            target_interface = device.make_lag(
                lag=changes.lag,
                ports=[local_port.name for local_port, remote_port in port_pairs],
                lag_min_links=changes.lag_links_min,
            )
            if changes.subif is not None:
                target_interface = device.add_subif(target_interface.name, changes.subif)
        elif changes.subif is not None:
            # single connection
            local_port, remote_port = port_pairs[0]
            target_interface = device.add_subif(local_port.name, changes.subif)
        elif changes.svi is not None:
            target_interface = device.add_svi(changes.svi)
        else:
            target_interface, _ = port_pairs[0]
        target_interface.add_addr(changes.addr, changes.vrf)
        return target_interface.name

    def _apply_nondirect_interface_changes(
            self, device: Device, ifname: Optional[str], changes: InterfaceChanges,
    ) -> Optional[str]:
        if changes.lag is not None:
            raise ValueError("LAG creation unsupported for indirect and virtual peers")
        elif changes.subif is not None:
            target_interface = device.add_subif(ifname, changes.subif)
        elif changes.svi is not None:
            target_interface = device.add_svi(changes.svi)
        elif not ifname:
            return None
        else:
            target_interface = device.find_interface(ifname)
            if not target_interface:
                raise ValueError(f"Interface {ifname} not found for device {device.fqdn}")
        target_interface.add_addr(changes.addr, changes.vrf)
        return target_interface.name

    def execute_for(self, device: Device) -> BgpConfig:
        all_fqdns = self._storage.resolve_all_fdnds()

        global_options = self._to_bgp_global(self._execute_globals(device))

        peers = []
        target_interface: Optional[str]
        for direct_pair in self._execute_direct(device):
            target_interface = self._apply_direct_interface_changes(
                device,
                direct_pair.device,
                direct_pair.ports,
                to_interface_changes(direct_pair.local, direct_pair.connected),
            )
            peers.append(self._to_bgp_peer(direct_pair, target_interface))

        for virtual_pair in self._execute_virtual(device):
            target_interface = self._apply_nondirect_interface_changes(
                device,
                getattr(virtual_pair.local, "ifname", None),
                to_interface_changes(virtual_pair.local, virtual_pair.connected),
            )
            peers.append(self._virtual_to_bgp_peer(virtual_pair, target_interface))

        for connected_pair in self._execute_indirect(device, all_fqdns):
            target_interface = self._apply_nondirect_interface_changes(
                device,
                getattr(connected_pair.local, "ifname", None),
                to_interface_changes(connected_pair.local, connected_pair.connected),
            )
            peers.append(self._to_bgp_peer(connected_pair, target_interface))

        return BgpConfig(
            global_options=global_options,
            peers=peers,
        )



from abc import abstractmethod
from typing import Protocol, Sequence

PortPair = tuple[str, str]


class PortProcessor(Protocol):
    @abstractmethod
    def __call__(self, pairs: Sequence[PortPair]) -> Sequence[list[PortPair]]:
        raise NotImplementedError


def united_ports(pairs: Sequence[PortPair]) -> Sequence[list[PortPair]]:
    return [list(pairs)]


def separate_ports(pairs: Sequence[PortPair]) -> Sequence[list[PortPair]]:
    return [[pair] for pair in pairs]



import sys
from abc import ABC, abstractmethod
from functools import cached_property
from importlib.metadata import entry_points
from typing import Generic, Optional, Type, TypeVar, List, Dict, Any, Tuple
import warnings
from annet.lib import get_context

T = TypeVar("T")


class Connector(ABC, Generic[T]):
    name: str
    # legacy
    ep_name: str
    ep_group: str = "annet.connectors"
    # right way just to use ep groups
    ep_by_group_only: str = ""
    _classes: Optional[List[Type[T]]] = None

    def _get_default(self) -> Type[T]:
        raise RuntimeError(f"{self.name} is not set")

    @cached_property
    def _entry_point(self) -> List[Type[T]]:
        ep = load_entry_point(self.ep_group, self.ep_name)
        if self.ep_by_group_only:
            ep.extend(load_entry_point_new(self.ep_by_group_only))
        return ep

    def get(self, *args, **kwargs) -> T:
        """
        Returns connector. If more than one is registered returns random and throw warning
        """
        if self._classes is None:
            self._classes = self._entry_point or [self._get_default()]
        if not self._classes:
            raise Exception(f"Not found registered class for group={self.ep_group}")
        if len(self._classes) > 1:
            warnings.warn(f"Multiple classes are registered with the group={self.ep_group} but "
                          f"{[cls for cls in self._classes]}", UserWarning)
        res = self._classes[0]
        return res(*args, **kwargs)

    def get_all(self) -> List[T]:
        if self._classes is None:
            self._classes = self._entry_point or [self._get_default()]

        return self._classes.copy()

    def set(self, cls: Type[T]):
        if self._classes is not None:
            raise RuntimeError(f"Cannot reinitialize value of {self.name}")
        self._classes = [cls]

    def set_all(self, classes: List[Type[T]]):
        if self._classes is not None:
            raise RuntimeError(f"Cannot reinitialize value of {self.name}")
        self._classes = list(classes)

    def is_default(self) -> bool:
        return self._classes is None and not self._entry_point


class CachedConnector(Connector[T], ABC):
    _cache: Optional[T] = None

    def get(self, *args, **kwargs) -> T:
        assert not (args or kwargs), "Arguments forwarding is not allowed for cached connectors"
        if self._cache is None:
            self._cache = super().get()
        return self._cache

    def set(self, cls: Type[T]):
        super().set(cls)
        self._cache = None


def load_entry_point(group: str, name: str):
    if sys.version_info < (3, 10):
        ep = [item for item in entry_points().get(group, []) if item.name == name]
    else:
        ep = entry_points(group=group, name=name)  # pylint: disable=unexpected-keyword-arg
    if not ep:
        return []
    return [item.load() for item in ep]


def load_entry_point_new(group: str) -> List:
    if sys.version_info < (3, 10):
        ep = [item for item in entry_points().get(group, [])]
    else:
        ep = entry_points(group=group)  # pylint: disable=unexpected-keyword-arg
    if not ep:
        return []
    return [item.load() for item in ep]


class AdapterWithConfig(ABC, Generic[T]):
    @classmethod
    @abstractmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> T:
        pass


class AdapterWithName(ABC):
    @classmethod
    @abstractmethod
    def name(cls) -> str:
        pass


def get_connector_from_config(config_key: str, connectors: List[Type[Connector]]) -> Tuple[Connector, Dict[str, Any]]:
    if not connectors:
        raise Exception("empty connectors")

    # handle configuration
    connector_params = dict[str, Any]()  # default
    if context_storage := get_context().get(config_key):
        connector_params = context_storage.get("params", {})
        if adapter_name := context_storage.get("adapter", None):
            seen = list[str]()
            for con in connectors:
                if issubclass(con, AdapterWithName):
                    con_name = con.name()
                else:
                    con_name = con.__name__
                seen.append(con_name)
                if adapter_name == con_name:
                    connectors = [con]
                    break
            else:
                raise Exception("unknown %s %s: seen %s" % (config_key, adapter_name, seen))

    if len(connectors) > 1:
        warnings.warn(f"Please specify adapter for '{config_key}'. Found more than one classes {connectors}", UserWarning)
    connector = connectors[0]
    if issubclass(connector, AdapterWithConfig):
        connector_ins = connector.with_config(**connector_params)
    else:
        connector_ins = connector()
    # return connector_params only for storage
    # TODO: switch storage interface to AdapterWithConfig
    return connector_ins, connector_params



import abc
from typing import ClassVar

from annet.annlib.command import CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import CommonFormatter


class AbstractVendor(abc.ABC):
    NAME: ClassVar[str]

    @abc.abstractmethod
    def match(self) -> list[str]:
        raise NotImplementedError

    def apply(
        self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str
    ) -> tuple[CommandList, CommandList]:
        return CommandList(), CommandList()

    @property
    @abc.abstractmethod
    def reverse(self) -> str:
        raise NotImplementedError

    def diff(self, order: bool) -> str:
        return "common.ordered_diff" if order else "common.default_diff"

    @property
    @abc.abstractmethod
    def exit(self) -> str:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def hardware(self) -> HardwareView:
        raise NotImplementedError

    def svi_name(self, num: int) -> str:
        return f"vlan{num}"

    @abc.abstractmethod
    def make_formatter(self, **kwargs) -> CommonFormatter:
        raise NotImplementedError




from typing import List

from annet.generators import BaseGenerator, PartialGenerator
from annet.storage import Storage


# ====
class Lldp(PartialGenerator):
    TAGS = ["mgmt", "lldp"]

    def acl_huawei(self, device):
        return """
            lldp
        """

    def run_huawei(self, device):
        yield "lldp enable"

        if device.hw.CE:
            yield "lldp transmit interval 10"

    def acl_nexus(self, device):
        return """
            feature lldp
            lldp
        """

    def run_nexus(self, device):
        yield "feature lldp"
        yield "lldp timer 10"

    def acl_juniper(self, _):
        return """
        protocols    %cant_delete
            lldp
                *
        """

    def run_juniper(self, device):
        with self.multiblock("protocols", "lldp"):
            yield """
                neighbour-port-info-display port-id
                port-description-type interface-alias
                port-id-subtype interface-name
                interface all
            """

    def acl_b4com(self, device):
        return """
        lldp *
        interface *
         lldp-agent
            *
        """

    def run_b4com(self, device):
        yield """
        lldp run
        lldp tlv-select basic-mgmt port-description
        lldp tlv-select basic-mgmt system-name
        lldp tlv-select basic-mgmt system-capabilities
        lldp tlv-select basic-mgmt system-description
        lldp tlv-select basic-mgmt management-address
        """
        for iface in device.interfaces:
            with self.multiblock(f"interface {iface.name}"):
                with self.multiblock("lldp-agent"):
                    yield """
                    set lldp enable txrx
                    set lldp chassis-id-tlv ip-address
                    set lldp port-id-tlv if-name
                    lldp tlv basic-mgmt system-name select
                    lldp tlv basic-mgmt system-description select
                    """


def get_generators(store: Storage) -> List[BaseGenerator]:
    return [
        Lldp(store),
    ]



from dataclasses import dataclass, field

from typing import List, Optional


FIRST_EXCEPTION = 1
ALL_COMPLETED = 2


@dataclass
class Question:
    question: str  # frame it using / if it is a regular expression
    answer: str
    is_regexp: Optional[bool] = False
    not_send_nl: bool = False


@dataclass
class Command:
    cmd: str | bytes
    questions: Optional[List[Question]] = None
    exc_handler: Optional[List[Question]] = None
    timeout: Optional[int] = None  # total timeout
    read_timeout: Optional[int] = None  # timeout between consecutive reads
    suppress_nonzero: bool = False
    suppress_eof: bool = False

    def __str__(self) -> str:
        if isinstance(self.cmd, bytes):
            return self.cmd.decode("utf-8")
        return self.cmd


@dataclass
class CommandList:
    cmss: List[Command] = field(default_factory=list)

    def __post_init__(self):
        if not self.cmss:
            self.cmss = []

    def __iter__(self):
        return iter(self.cmss)

    def __len__(self) -> int:
        return len(self.cmss)

    def add_cmd(self, cmd: Command) -> None:
        assert isinstance(cmd, Command)
        self.cmss.append(cmd)

    def as_list(self) -> List[Command]:  # TODO: delete
        return self.cmss



from logging import getLogger

from annet.annlib.netdev.views.hardware import HardwareView

logger = getLogger(__name__)


def get_hw(manufacturer: str, model: str, platform_name: str):
    # By some reason Netbox calls Mellanox SN as MSN, so we fix them here
    if manufacturer == "Mellanox" and model.startswith("MSN"):
        model = model.replace("MSN", "SN", 1)

    return HardwareView(manufacturer + " " + model, platform_name)


def get_breed(manufacturer: str, model: str):
    hw = get_hw(manufacturer, model, "")
    if hw.Huawei.CE:
        return "vrp85"
    elif hw.Huawei.NE:
        return "vrp85"
    elif hw.Huawei:
        return "vrp55"
    elif hw.H3C:
        return "h3c"
    elif hw.PC.NVIDIA or hw.PC.Mellanox:
        return "cuml2"
    elif hw.Juniper:
        return "jun10"
    elif hw.Cisco.Nexus:
        return "nxos"
    elif hw.Cisco:
        return "ios12"
    elif hw.Arista:
        return "eos4"
    elif hw.B4com:
        return "bcom-os"
    elif hw.RouterOS:
        return "routeros"
    elif hw.PC.Moxa or hw.PC.Nebius:
        return "moxa"
    elif hw.PC:
        return "pc"
    return ""



from abc import abstractmethod, ABC
from typing import Generic, TypeVar

from annet.annlib.netdev.views.hardware import HardwareView
from .manufacturer import get_breed, get_hw
from .models import NetboxDevice, Interface, IpAddress, Prefix, \
    FHRPGroupAssignment, FHRPGroup

NetboxDeviceT = TypeVar("NetboxDeviceT", bound=NetboxDevice)
InterfaceT = TypeVar("InterfaceT", bound=Interface)
IpAddressT = TypeVar("IpAddressT", bound=IpAddress)
PrefixT = TypeVar("PrefixT", bound=Prefix)
FHRPGroupT = TypeVar("FHRPGroupT", bound=FHRPGroup)
FHRPGroupAssignmentT = TypeVar(
    "FHRPGroupAssignmentT", bound=FHRPGroupAssignment,
)


def get_device_breed(device: NetboxDeviceT) -> str:
    if device.device_type and device.device_type.manufacturer:
        return get_breed(
            device.device_type.manufacturer.name,
            device.device_type.model,
        )
    return ""


def get_device_hw(device: NetboxDeviceT) -> HardwareView:
    if device.device_type and device.device_type.manufacturer:
        return get_hw(
            device.device_type.manufacturer.name,
            device.device_type.model,
            device.platform.name if device.platform else "",
        )
    return HardwareView("", "")


class NetboxAdapter(
    ABC,
    Generic[
        NetboxDeviceT,
        InterfaceT,
        IpAddressT,
        PrefixT,
        FHRPGroupT,
        FHRPGroupAssignmentT,
    ],
):
    @abstractmethod
    def list_fqdns(self, query: dict[str, list[str]] | None = None) -> list[str]:
        raise NotImplementedError()

    @abstractmethod
    def list_devices(self, query: dict[str, list[str]]) -> list[NetboxDeviceT]:
        raise NotImplementedError()

    @abstractmethod
    def get_device(self, device_id: int) -> NetboxDeviceT:
        raise NotImplementedError()

    @abstractmethod
    def list_interfaces_by_devices(self, device_ids: list[int]) -> list[InterfaceT]:
        raise NotImplementedError()

    @abstractmethod
    def list_interfaces(self, ids: list[int]) -> list[InterfaceT]:
        raise NotImplementedError()

    @abstractmethod
    def list_ipaddr_by_ifaces(self, iface_ids: list[int]) -> list[IpAddressT]:
        raise NotImplementedError()

    @abstractmethod
    def list_ipprefixes(self, prefixes: list[str]) -> list[PrefixT]:
        raise NotImplementedError()

    @abstractmethod
    def list_fhrp_group_assignments(
            self, iface_ids: list[int],
    ) -> list[FHRPGroupAssignmentT]:
        raise NotImplementedError()

    @abstractmethod
    def list_fhrp_groups(
            self, ids: list[int],
    ) -> list[FHRPGroupT]:
        raise NotImplementedError()



import re
from abc import ABC
from os import path
from typing import Iterable, Union

from annet.annlib.lib import mako_render
from annet.annlib.rbparser.ordering import compile_ordering_text
from annet.annlib.rbparser.platform import VENDOR_ALIASES

from annet.connectors import CachedConnector
from annet.rulebook.deploying import compile_deploying_text
from annet.rulebook.patching import compile_patching_text
from annet.vendors import registry_connector


class RulebookProvider(ABC):
    def get_rulebook(self, hw):
        raise NotImplementedError

    def get_root_modules(self) -> Iterable[str]:
        raise NotImplementedError


class _RulebookProviderConnector(CachedConnector[RulebookProvider]):
    name = "Rulebook provider"
    ep_name = "rulebook"


rulebook_provider_connector = _RulebookProviderConnector()


def get_rulebook(hw):
    return rulebook_provider_connector.get().get_rulebook(hw)


class DefaultRulebookProvider(RulebookProvider):
    root_dir = (path.dirname(__file__),)
    root_modules = ("annet.rulebook",)

    def __init__(self, root_dir: Union[str, Iterable[str], None] = None,
                 root_modules: Union[str, Iterable[str], None] = None):
        self._rulebook_cache = {}
        self._render_rul_cache = {}
        self._escaped_rul_cache = {}

        if root_dir is None:
            pass
        elif isinstance(root_dir, str):
            self.root_dir = (root_dir,)
        else:
            self.root_dir = tuple(root_dir)

        if root_modules is None:
            pass
        elif isinstance(root_modules, str):
            self.root_modules = (root_modules,)
        else:
            self.root_modules = tuple(root_modules)

    def get_root_modules(self):
        return self.root_modules

    def get_rulebook(self, hw):
        if hw in self._rulebook_cache:
            return self._rulebook_cache[hw]

        assert hw.vendor in registry_connector.get(), "Unknown vendor: %s" % (hw.vendor)
        rul_vendor_name = VENDOR_ALIASES.get(hw.vendor, hw.vendor)
        patching = compile_patching_text(self._render_rul(rul_vendor_name + ".rul", hw), rul_vendor_name)

        try:
            ordering_text = self._render_rul(hw.vendor + ".order", hw)
        except FileNotFoundError:
            ordering_text = ""
        ordering = compile_ordering_text(ordering_text, hw.vendor)

        try:
            deploying_text = self._render_rul(hw.vendor + ".deploy", hw)
        except FileNotFoundError:
            deploying_text = ""

        deploying = compile_deploying_text(deploying_text, hw.vendor)

        self._rulebook_cache[hw] = {
            "patching": patching,
            "ordering": ordering,
            "deploying": deploying,
        }
        return self._rulebook_cache[hw]

    def _render_rul(self, name, hw):
        key = (name, hw)
        if key not in self._render_rul_cache:
            self._render_rul_cache[key] = mako_render(self._read_escaped_rul(name), hw=hw)
        return self._render_rul_cache[key]

    def _read_escaped_rul(self, name):
        if name in self._escaped_rul_cache:
            return self._escaped_rul_cache[name]
        for root_dir in self.root_dir:
            try:
                with open(path.join(root_dir, "texts", name), "r", encoding="utf-8") as f:
                    self._escaped_rul_cache[name] = self._escape_mako(f.read())
                    return self._escaped_rul_cache[name]
            except FileNotFoundError:
                pass

        raise FileNotFoundError(f"Unable to find rul: {name}")

    @staticmethod
    def _escape_mako(text):
        # Экранирование всего, что начинается на %, например %comment -> %%comment, чтобы он не интерпретировался
        # как mako-оператор
        text = re.sub(r"(?:^|\n)%((?!if\s*|elif\s*|else\s*|endif\s*|for\s*|endfor\s*))", "\n%%\\1", text)
        text = re.sub(r"(?:^|\n)\s*#.*", "", text)
        return text



from annet.annlib.types import Op

from annet.rulebook import common


def remove_last_param(rule, key, diff, **_):
    if diff[Op.REMOVED]:
        for rem in diff[Op.REMOVED]:
            # Обрабатывать удаление последнего параметра команды
            cmd_parts = rem["row"].split(" ")
            cmd_parts.remove(cmd_parts[len(cmd_parts) - 1])
            yield False, "undo %s" % " ".join(cmd_parts), None
    else:
        yield from common.default(rule, key, diff)



from typing import Dict, Any, Optional

from dataclass_rest.exceptions import ClientError, ClientLibraryError

from annet.storage import StorageProvider, Storage
from annet.connectors import AdapterWithName, AdapterWithConfig, T
from .common.status_client import NetboxStatusClient
from .common.storage_opts import NetboxStorageOpts
from .common.query import NetboxQuery
from .v24.storage import NetboxStorageV24
from .v37.storage import NetboxStorageV37
from .v41.storage import NetboxStorageV41
from .v42.storage import NetboxStorageV42


def storage_factory(opts: NetboxStorageOpts) -> Storage:
    client = NetboxStatusClient(opts.url, opts.token, opts.insecure)
    version_class_map = {
        "3.4": NetboxStorageV37,
        "3.7": NetboxStorageV37,
        "4.0": NetboxStorageV41,
        "4.1": NetboxStorageV41,
        "4.2": NetboxStorageV42,
        "4.3": NetboxStorageV42,
    }

    status = None

    try:
        status = client.status()
        for version_prefix, storage_class in version_class_map.items():
            if version_prefix == status.minor_version:
                return storage_class(opts)

    except ClientError as e:
        if e.status_code == 404:
            return NetboxStorageV24(opts)
        else:
            raise ValueError(f"Unsupported version: {status.netbox_version}")
    except ClientLibraryError:
        raise ValueError(
            f"Connection error: Unable to reach Netbox at URL: {opts.url}")
    raise Exception(f"Unsupported version: {status.netbox_version}")


class NetboxProvider(StorageProvider, AdapterWithName, AdapterWithConfig):
    def __init__(self, url: Optional[str] = None, token: Optional[str] = None, insecure: bool = False,
                 exact_host_filter: bool = False, threads: int = 1, all_hosts_filter: dict[str, list[str]] | None = None):
        self.url = url
        self.token = token
        self.insecure = insecure
        self.exact_host_filter = exact_host_filter
        self.threads = threads
        self.all_hosts_filter = all_hosts_filter

    @classmethod
    def with_config(cls, **kwargs: Dict[str, Any]) -> T:
        return cls(**kwargs)

    def storage(self):
        return storage_factory

    def opts(self):
        return NetboxStorageOpts

    def query(self):
        return NetboxQuery

    @classmethod
    def name(cls) -> str:
        return "netbox"



from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import OptixtransFormatter
from annet.vendors.registry import registry

from .huawei import HuaweiVendor


@registry.register
class OptixTransVendor(HuaweiVendor):
    NAME = "optixtrans"

    def match(self) -> list[str]:
        return ["Huawei.OptiXtrans"]

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("Huawei OptiXtrans")

    def make_formatter(self, **kwargs) -> OptixtransFormatter:
        return OptixtransFormatter(**kwargs)

    def svi_name(self, num: int) -> str:
        return f"vlan{num}"



import ssl

from adaptix import P
from adaptix.conversion import get_converter, link_function, link_constant, link
from annetbox.v37 import models as api_models, client_sync

from annet.adapters.netbox.common.adapter import NetboxAdapter, get_device_breed, get_device_hw
from annet.adapters.netbox.common.storage_base import BaseNetboxStorage
from annet.storage import Storage
from .models import (
    IpAddressV37, NetboxDeviceV37, InterfaceV37, PrefixV37,
    FHRPGroupAssignmentV37, FHRPGroupV37,
)


class NetboxV37Adapter(NetboxAdapter[
    NetboxDeviceV37, InterfaceV37, IpAddressV37, PrefixV37, FHRPGroupV37, FHRPGroupAssignmentV37,
]):
    def __init__(
            self,
            storage: Storage,
            url: str,
            token: str,
            ssl_context: ssl.SSLContext | None,
            threads: int,
    ):
        self.netbox = client_sync.NetboxV37(url=url, token=token, ssl_context=ssl_context, threads=threads)
        self.convert_device = get_converter(
            api_models.Device,
            NetboxDeviceV37,
            recipe=[
                link_function(get_device_breed, P[NetboxDeviceV37].breed),
                link_function(get_device_hw, P[NetboxDeviceV37].hw),
                link_constant(P[NetboxDeviceV37].interfaces, factory=list),
                link_constant(P[NetboxDeviceV37].storage, value=storage),
                link(P[api_models.Device].name, P[NetboxDeviceV37].hostname),
                link(P[api_models.Device].name, P[NetboxDeviceV37].fqdn),
            ]
        )
        self.convert_interfaces = get_converter(
            list[api_models.Interface],
            list[InterfaceV37],
            recipe=[
                link_constant(P[InterfaceV37].ip_addresses, factory=list),
                link_constant(P[InterfaceV37].fhrp_groups, factory=list),
                link_constant(P[InterfaceV37].lag_min_links, value=None),
            ]
        )
        self.convert_ip_addresses = get_converter(
            list[api_models.IpAddress],
            list[IpAddressV37],
            recipe=[
                link_constant(P[IpAddressV37].prefix, value=None),
            ]
        )
        self.convert_ip_prefixes = get_converter(
            list[api_models.Prefix],
            list[PrefixV37],
        )
        self.convert_fhrp_group_assignments = get_converter(
            list[api_models.FHRPGroupAssignmentBrief],
            list[FHRPGroupAssignmentV37],
            recipe=[
                link_constant(P[FHRPGroupAssignmentV37].group, value=None),
                link_function(lambda model: model.group.id, P[FHRPGroupAssignmentV37].fhrp_group_id),
            ]
        )
        self.convert_fhrp_groups = get_converter(
            list[api_models.FHRPGroup],
            list[FHRPGroupV37],
        )

    def list_fqdns(self, query: dict[str, list[str]] | None = None) -> list[str]:
        query = query or {}
        return [
            d.name
            for d in self.netbox.dcim_all_devices_brief(**query).results
        ]

    def list_devices(self, query: dict[str, list[str]]) -> list[NetboxDeviceV37]:
        return [
            self.convert_device(dev)
            for dev in self.netbox.dcim_all_devices(**query).results
        ]

    def get_device(self, device_id: int) -> NetboxDeviceV37:
        return self.convert_device(self.netbox.dcim_device(device_id))

    def list_interfaces_by_devices(self, device_ids: list[int]) -> list[InterfaceV37]:
        return self.convert_interfaces(self.netbox.dcim_all_interfaces(device_id=device_ids).results)

    def list_interfaces(self, ids: list[int]) -> list[InterfaceV37]:
        return self.convert_interfaces(self.netbox.dcim_all_interfaces(id=ids).results)

    def list_ipaddr_by_ifaces(self, iface_ids: list[int]) -> list[IpAddressV37]:
        return self.convert_ip_addresses(self.netbox.ipam_all_ip_addresses(interface_id=iface_ids).results)

    def list_ipprefixes(self, prefixes: list[str]) -> list[PrefixV37]:
        return self.convert_ip_prefixes(self.netbox.ipam_all_prefixes(prefix=prefixes).results)

    def list_fhrp_group_assignments(
            self, iface_ids: list[int],
    ) -> list[FHRPGroupAssignmentV37]:
        raw_assignments = self.netbox.ipam_all_fhrp_group_assignments_by_interface(
            interface_id=iface_ids,
        )
        return self.convert_fhrp_group_assignments(raw_assignments.results)

    def list_fhrp_groups(self, ids: list[int]) -> list[FHRPGroupV37]:
        raw_groups = self.netbox.ipam_all_fhrp_groups_by_id(id=list(ids))
        return self.convert_fhrp_groups(raw_groups.results)


class NetboxStorageV37(BaseNetboxStorage[
    NetboxDeviceV37, InterfaceV37, IpAddressV37, PrefixV37, FHRPGroupV37, FHRPGroupAssignmentV37,
]):
    def _init_adapter(
            self,
            url: str,
            token: str,
            ssl_context: ssl.SSLContext | None,
            threads: int,
    ) -> NetboxAdapter[NetboxDeviceV37, InterfaceV37, IpAddressV37, PrefixV37, FHRPGroupV37, FHRPGroupAssignmentV37]:
        return NetboxV37Adapter(self, url, token, ssl_context, threads)



from typing import List

from annet.adapters.netbox.common.models import NetboxDevice
from annet.generators import PartialGenerator, BaseGenerator
from annet.mesh import MeshExecutor
from annet.storage import Storage
from .mesh_logic import registry


class Bgp(PartialGenerator):
    TAGS = ["mgmt", "bgp"]

    def acl_huawei(self, device):
        return """
        bgp
            peer
        """

    def run_huawei(self, device: NetboxDevice):
        executor = MeshExecutor(registry, device.storage)
        res = executor.execute_for(device)
        yield f"bgp {res.global_options.local_as}"
        for peer in res.peers:
            yield f"   peer {peer.addr} interface {peer.interface}"
            if peer.group_name:
                yield f"   peer {peer.addr} {peer.group_name}"
            if peer.remote_as is not None:
                yield f"   peer {peer.addr} remote-as {peer.remote_as}"

        for group in res.global_options.groups:
            yield f"   peer {group.name} remote-as {group.remote_as}"
        for interface in device.interfaces:
            print(
                interface.name,
                interface.lag_min_links,
                interface.lag.name if interface.lag else None,
                interface.ip_addresses,
            )


def get_generators(store: Storage) -> List[BaseGenerator]:
    return [
        Bgp(store),
    ]



import warnings
from collections.abc import Callable
from dataclasses import dataclass
from dataclasses import field
from enum import Enum
from typing import Optional, Literal, TypeVar, Union, Any

from .action import SingleAction, ActionType
from .policy import RoutingPolicyStatement
from .result import ResultType


class ThenField(str, Enum):
    community = "community"
    large_community = "large_community"
    extcommunity_rt = "extcommunity_rt"
    extcommunity_soo = "extcommunity_soo"
    extcommunity = "extcommunity"
    as_path = "as_path"
    local_pref = "local_pref"
    metric = "metric"
    rpki_valid_state = "rpki_valid_state"
    resolution = "resolution"
    mpls_label = "mpls_label"
    metric_type = "metric_type"
    origin = "origin"
    tag = "tag"

    next_hop = "next_hop"


ValueT = TypeVar("ValueT")
_Setter = Callable[[ValueT], SingleAction[ValueT]]


@dataclass
class CommunityActionValue:
    replaced: Optional[list[str]] = None  # None means no replacement is done
    added: list[str] = field(default_factory=list)
    removed: list[str] = field(default_factory=list)

    def __bool__(self) -> bool:  # check if any action required
        return bool(self.replaced is not None or self.added or self.removed)


class CommunityActionBuilder:
    def __init__(self, community: CommunityActionValue):
        self._community = community

    def add(self, *community: str) -> None:
        for c in community:
            self._community.added.append(c)

    def remove(self, *community: str) -> None:
        for c in community:
            self._community.removed.append(c)

    def set(self, *community: str) -> None:
        self._community.added.clear()
        self._community.removed.clear()
        self._community.replaced = list(community)


@dataclass
class AsPathActionValue:
    set: Optional[list[str]] = None  # None means no replacement is done
    prepend: list[str] = field(default_factory=list)
    expand: list[str] = field(default_factory=list)
    expand_last_as: str = ""
    delete: list[str] = field(default_factory=list)

    def __bool__(self) -> bool:  # check if any action required
        return bool(
            self.set is not None or self.prepend or self.expand or self.expand_last_as or self.delete
        )


RawAsNum = Union[str, int]


class AsPathActionBuilder:
    def __init__(self, as_path_value: AsPathActionValue):
        self._as_path_value = as_path_value

    def prepend(self, *values: RawAsNum) -> None:
        self._as_path_value.prepend = list(map(str, values))

    def delete(self, *values: RawAsNum) -> None:
        self._as_path_value.delete = list(map(str, values))

    def expand(self, *values: RawAsNum) -> None:
        self._as_path_value.expand = list(map(str, values))

    def expand_last_as(self, value: RawAsNum) -> None:
        self._as_path_value.expand_last_as = str(value)

    def set(self, *values: RawAsNum) -> None:
        self._as_path_value.expand.clear()
        self._as_path_value.expand_last_as = ""
        self._as_path_value.delete.clear()
        self._as_path_value.prepend.clear()
        self._as_path_value.set = list(map(str, values))


@dataclass
class NextHopActionValue:
    target: Optional[Literal["self", "discard", "peer", "ipv4_addr", "ipv6_addr", "mapped_ipv4"]] = None
    addr: str = ""

    def __bool__(self) -> bool:
        return bool(self.target)


class NextHopActionBuilder:
    def __init__(self, next_hop_value: NextHopActionValue):
        self._next_hop_value = next_hop_value

    def ipv4_addr(self, value: str) -> None:
        self._next_hop_value.target = "ipv4_addr"
        self._next_hop_value.addr = value

    def ipv6_addr(self, value: str) -> None:
        self._next_hop_value.target = "ipv6_addr"
        self._next_hop_value.addr = value

    def mapped_ipv4(self, value: str) -> None:
        self._next_hop_value.target = "mapped_ipv4"
        self._next_hop_value.addr = value

    def self(self) -> None:
        self._next_hop_value.target = "self"
        self._next_hop_value.addr = ""

    def peer(self) -> None:
        self._next_hop_value.target = "peer"
        self._next_hop_value.addr = ""

    def discard(self) -> None:
        self._next_hop_value.target = "discard"
        self._next_hop_value.addr = ""


class StatementBuilder:
    def __init__(self, statement: RoutingPolicyStatement) -> None:
        self._statement = statement
        self._added_as_path: list[int] = []
        self._community = CommunityActionValue()
        self._large_community = CommunityActionValue()
        self._extcommunity = CommunityActionValue()
        self._extcommunity_rt = CommunityActionValue()
        self._extcommunity_soo = CommunityActionValue()
        self._as_path = AsPathActionValue()
        self._next_hop = NextHopActionValue()

    @property
    def next_hop(self) -> NextHopActionBuilder:
        return NextHopActionBuilder(self._next_hop)

    @property
    def as_path(self) -> AsPathActionBuilder:
        return AsPathActionBuilder(self._as_path)

    @property
    def community(self) -> CommunityActionBuilder:
        return CommunityActionBuilder(self._community)

    @property
    def large_community(self) -> CommunityActionBuilder:
        return CommunityActionBuilder(self._large_community)

    @property
    def extcommunity(self) -> CommunityActionBuilder:
        return CommunityActionBuilder(self._extcommunity)

    @property
    def extcommunity_rt(self) -> CommunityActionBuilder:
        warnings.warn("extcommunity_rt is deprecated, use extcommunity", DeprecationWarning, stacklevel=2)
        return CommunityActionBuilder(self._extcommunity_rt)

    @property
    def extcommunity_soo(self) -> CommunityActionBuilder:
        warnings.warn("extcommunity_soo is deprecated, use extcommunity", DeprecationWarning, stacklevel=2)
        return CommunityActionBuilder(self._extcommunity_soo)

    def _set(self, field: str, value: ValueT) -> None:
        action = self._statement.then
        if field in action:
            action[field].type = ActionType.SET
            action[field].value = value
        else:
            action.append(SingleAction(
                field=field,
                type=ActionType.SET,
                value=value,
            ))

    def set_local_pref(self, value: int) -> None:
        self._set(ThenField.local_pref, value)

    def set_metric_type(self, value: str) -> None:
        self._set(ThenField.metric_type, value)

    def set_metric(self, value: int) -> None:
        self._set(ThenField.metric, value)

    def add_metric(self, value: int) -> None:
        action = self._statement.then
        field = ThenField.metric
        if field in action:
            old_action = action[field]
            if old_action.type == ActionType.SET:
                action[field].value += value
            elif old_action.type == ActionType.ADD:
                action[field].value = value
            else:
                raise RuntimeError(f"Unknown action type {old_action.type} for metric")
        else:
            action.append(SingleAction(
                field=field,
                type=ActionType.ADD,
                value=value,
            ))

    def set_rpki_valid_state(self, value: str) -> None:
        self._set(ThenField.rpki_valid_state, value)

    def set_resolution(self, value: str) -> None:
        self._set(ThenField.resolution, value)

    def set_mpls_label(self) -> None:
        self._set(ThenField.mpls_label, True)

    def set_origin(self, value: str) -> None:
        self._set(ThenField.origin, value)

    def set_tag(self, value: int) -> None:
        self._set(ThenField.tag, value)

    def set_next_hop(self, value: Literal["self", "peer"]) -> None:  # ???
        self._set(ThenField.next_hop, value)

    def __enter__(self) -> "StatementBuilder":
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._community:
            self._statement.then.append(SingleAction(
                field=ThenField.community,
                type=ActionType.CUSTOM,
                value=self._community,
            ))
        if self._large_community:
            self._statement.then.append(SingleAction(
                field=ThenField.large_community,
                type=ActionType.CUSTOM,
                value=self._large_community,
            ))
        if self._extcommunity:
            self._statement.then.append(SingleAction(
                field=ThenField.extcommunity,
                type=ActionType.CUSTOM,
                value=self._extcommunity,
            ))
        if self._extcommunity_rt:
            self._statement.then.append(SingleAction(
                field=ThenField.extcommunity_rt,
                type=ActionType.CUSTOM,
                value=self._extcommunity_rt,
            ))
        if self._extcommunity_soo:
            self._statement.then.append(SingleAction(
                field=ThenField.extcommunity_soo,
                type=ActionType.CUSTOM,
                value=self._extcommunity_soo,
            ))
        if self._as_path:
            self._statement.then.append(SingleAction(
                field=ThenField.as_path,
                type=ActionType.CUSTOM,
                value=self._as_path,
            ))
        if self._next_hop:
            self._statement.then.append(SingleAction(
                field=ThenField.next_hop,
                type=ActionType.CUSTOM,
                value=self._next_hop,
            ))
        return None

    def allow(self) -> None:
        self._statement.result = ResultType.ALLOW

    def deny(self) -> None:
        self._statement.result = ResultType.DENY

    def next(self) -> None:
        self._statement.result = ResultType.NEXT

    def next_policy(self) -> None:
        self._statement.result = ResultType.NEXT_POLICY

    def add_as_path(self, *as_path: int) -> None:
        self._added_as_path.extend(as_path)

    def custom_action(self, action: SingleAction[Any]) -> None:
        self._statement.then.append(action)



import os
from typing import Any, Optional
from .query import parse_query

DEFAULT_URL = "http://localhost"


class NetboxStorageOpts:
    def __init__(
            self,
            url: str,
            token: str,
            insecure: bool = False,
            exact_host_filter: bool = False,
            threads: int = 1,
            all_hosts_filter: dict[str, list[str]] | None = None,
    ):
        self.url = url
        self.token = token
        self.insecure = insecure
        self.exact_host_filter = exact_host_filter
        self.threads = threads
        self.all_hosts_filter: dict[str, list[str]] = all_hosts_filter or {}

    @classmethod
    def parse_params(cls, conf_params: Optional[dict[str, str]], cli_opts: Any):
        url = os.getenv("NETBOX_URL") or conf_params.get("url") or DEFAULT_URL
        token = os.getenv("NETBOX_TOKEN", "").strip() or conf_params.get("token") or ""
        all_hosts_filter = None
        if all_hosts_filter_env := os.getenv("NETBOX_ALL_HOSTS_FILTER", "").strip():
            all_hosts_filter = parse_query(all_hosts_filter_env.split(","))
        elif all_hosts_filter_params := conf_params.get("all_hosts_filter"):
            all_hosts_filter = all_hosts_filter_params
        threads = os.getenv("NETBOX_CLIENT_THREADS", "").strip() or conf_params.get("threads") or "1"
        insecure = False
        if insecure_env := os.getenv("NETBOX_INSECURE", "").lower():
            insecure = insecure_env in ("true", "1", "t")
        else:
            insecure = bool(conf_params.get("insecure") or False)
        if exact_host_filter_env := os.getenv("NETBOX_EXACT_HOST_FILTER", "").lower():
            exact_host_filter = exact_host_filter_env in ("true", "1", "t")
        else:
            exact_host_filter = bool(conf_params.get("exact_host_filter") or False)
        return cls(
            url=url,
            token=token,
            insecure=insecure,
            exact_host_filter=exact_host_filter,
            threads=int(threads),
            all_hosts_filter=all_hosts_filter,
        )



from annet.generators import BaseGenerator
from annet.storage import Storage
from . import generator


def get_generators(storage: Storage) -> list[BaseGenerator]:
    return generator.get_generators(storage)



from dataclasses import dataclass
from typing import Optional
import warnings
from datetime import datetime, timezone
from annet.adapters.netbox.common.models import Entity, Interface, \
    InterfaceType, IpAddress, Label, NetboxDevice, DeviceIp, IpFamily, Prefix, \
    FHRPGroupAssignment, FHRPGroup, EntityWithSlug


@dataclass
class PrefixV41(Prefix):
    site: Optional[Entity] = None


@dataclass
class IpAddressV41(IpAddress[PrefixV41]):
    role: Optional[Label] = None
    prefix: Optional[PrefixV41] = None


@dataclass
class DeviceIpV41(DeviceIp):
    family: IpFamily


@dataclass
class FHRPGroupV41(FHRPGroup[DeviceIpV41]):
    pass


@dataclass
class FHRPGroupAssignmentV41(FHRPGroupAssignment[FHRPGroupV41]):
    pass


@dataclass
class InterfaceV41(Interface[IpAddressV41, FHRPGroupAssignmentV41]):
    def _add_new_addr(self, address_mask: str, vrf: Entity | None, family: IpFamily) -> None:
        self.ip_addresses.append(IpAddressV41(
            id=0,
            display=address_mask,
            address=address_mask,
            vrf=vrf,
            prefix=None,
            family=family,
            created=datetime.now(timezone.utc),
            last_updated=datetime.now(timezone.utc),
            tags=[],
            status=Label(value="active", label="Active"),
            assigned_object_id=self.id,
        ))


@dataclass
class NetboxDeviceV41(NetboxDevice[InterfaceV41, DeviceIpV41]):
    role: EntityWithSlug

    @property
    def device_role(self):
        warnings.warn(
            "'device_role' is deprecated, use 'role' instead.",
            DeprecationWarning,
            stacklevel=2
        )
        return self.role

    def __hash__(self):
        return hash((self.id, type(self)))

    def _make_interface(self, name: str, type: InterfaceType) -> InterfaceV41:
        return InterfaceV41(
            name=name,
            device=self,
            enabled=True,
            description="",
            type=type,
            id=0,
            vrf=None,
            display=name,
            untagged_vlan=None,
            tagged_vlans=[],
            ip_addresses=[],
            connected_endpoints=[],
            mode=None,
        )



from annet.annlib.command import Command, CommandList
from annet.annlib.netdev.views.hardware import HardwareView
from annet.vendors.tabparser import RibbonFormatter
from annet.vendors.base import AbstractVendor
from annet.vendors.registry import registry


@registry.register
class RibbonVendor(AbstractVendor):
    NAME = "ribbon"

    def apply(self, hw: HardwareView, do_commit: bool, do_finalize: bool, path: str) -> tuple[CommandList, CommandList]:
        before, after = CommandList(), CommandList()

        before.add_cmd(Command("configure exclusive"))
        if do_commit:
            after.add_cmd(Command("commit", timeout=30))
        after.add_cmd(Command("exit"))

        return before, after

    def match(self) -> list[str]:
        return ["Ribbon"]

    @property
    def reverse(self) -> str:
        return "delete"

    @property
    def hardware(self) -> HardwareView:
        return HardwareView("Ribbon")

    def make_formatter(self, **kwargs) -> RibbonFormatter:
        return RibbonFormatter(**kwargs)

    @property
    def exit(self) -> str:
        return "exit"

    def diff(self, order: bool) -> str:
        return "juniper.ordered_diff" if order else "juniper.default_diff"



from typing import Callable

from annet.connectors import Connector
from annet.vendors.base import AbstractVendor

from .library import (
    arista,
    aruba,
    b4com,
    cisco,
    h3c,
    huawei,
    iosxr,
    juniper,
    nexus,
    nokia,
    optixtrans,
    pc,
    ribbon,
    routeros,
)
from .registry import Registry, registry


class _RegistryConnector(Connector[Registry]):
    name = "Registry"
    ep_name = "vendors"

    def _get_default(self) -> Callable[[], Registry]:
        return lambda: registry


registry_connector = _RegistryConnector()



import socket

from annet.annlib.types import Op

from annet.rulebook import common


def undo_commit(rule, key, diff, **_):
    # Huawei не даёт снести конфигурацию bgp и написать заново одним коммитом. Говорит:
    #    Invalid configuration. BGP is under undo.
    # при попытке создать новую после удаления
    if diff[Op.REMOVED]:
        rule["force_commit"] = True
        yield (False, rule["reverse"], None)
    # commit нужен под undo bgp
    rule["force_commit"] = False
    yield from common.default(rule, key, diff)


def peer(rule, key, diff, **_):  # pylint: disable=unused-argument
    """
        Особенность peer-команд в том, что
            peer IP as-number N
        является основной командой, и отменить её можно только через
            undo peer IP
        , то есть полностью удалив все настройки пира.

        При этом, as-number может выставляться и для группы:
            group SPINES
            peer SPINES as-number 13238
        в таком случае игнорим, позволяем удалить эту настройку поскольку она не дефайнит группу
            undo peer SPINES as-number
    """

    assert not diff[Op.AFFECTED], "Peer commands could not contain subcommands"
    for action in sorted(diff[Op.REMOVED], key=lambda act: "as-number" not in act["row"].split()):
        tokens = action["row"].split()
        (_, addr_or_group_name, param, *__) = tokens
        if param == "as-number":
            if _is_ip_addr(addr_or_group_name):
                yield (False, "undo peer {}".format(*key), None)
            else:
                # мы не можем делать common.default потому что правило определено как peer * а не peer * *
                # таким образом дефолтное поведение тут будет "undo peer PEERGROUP" что не то что мы хотим
                yield (False, "undo peer {} as-number".format(*key), None)
            break

        if param in ["connect-interface", "ebgp-max-hop", "local-as", "substitute-as", "password", "preferred-value"]:
            yield (False, "undo " + " ".join(tokens[:3]), None)
        else:
            yield (False, "undo " + action["row"], None)

    for action in sorted(diff[Op.ADDED], key=lambda act: "as-number" not in act["row"]):
        yield (True, action["row"], None)


def bfd(rule, key, diff, **_):
    """
    [*vla-1x1-bgp]undo peer SPINE1 bfd min-tx-interval 500 min-rx-interval 500 detect-multiplier 4
    │Error: Unrecognized command found at '^' position.

    [*vla-1x1-bgp]undo peer SPINE1 bfd min-rx-interval
    [~vla-1x1-bgp]undo peer SPINE1 bfd min-tx-interval
    [*vla-1x1-bgp]undo peer SPINE1 bfd detect-multiplier
    """
    if diff[Op.REMOVED]:
        assert len(diff[Op.REMOVED]) <= 1 and len(diff[Op.ADDED]) <= 1
        new_params = set()
        if diff[Op.ADDED]:
            new_params = set(_bfd_params_used(diff[Op.ADDED][0]["row"]))
        for token in _bfd_params_used(diff[Op.REMOVED][0]["row"]):
            if token not in new_params:
                yield (False, rule["reverse"].format(*key) + " " + token, None)
        diff[Op.REMOVED] = []
    if diff[Op.ADDED]:
        yield from common.default(rule, key, diff, **_)


def _is_ip_addr(addr_or_string):
    ret = None
    for af in (socket.AF_INET6, socket.AF_INET):
        try:
            ret = socket.inet_pton(af, addr_or_string)
        except OSError:
            pass
        else:
            break
    return bool(ret)


def _bfd_params_used(row):
    prev = None
    for token in row.split():
        if prev and token.isnumeric():
            if prev and token.isnumeric():
                yield prev
        prev = token



import abc
import os
import posixpath
import sys
from typing import Dict, List, Optional, Tuple, Type
from urllib.parse import urlparse

import colorama
from contextlog import get_logger

from annet.annlib.output import (  # pylint: disable=unused-import
    LABEL_NEW_PREFIX,
    OutputWriter,
    TextArgs,
    capture_output,
    dir_or_file_output,
    format_file_diff,
    print_as_json,
    print_as_yaml,
    print_err_label,
    print_label,
)
from annet.cli_args import FileOutOptions, QueryOptions
from annet.connectors import Connector
from annet.storage import Device, storage_connector


BLACKBOX_FILENAME = "config.cfg"


class _DriverConnector(Connector["OutputDriver"]):
    name = "OutputDriver"
    ep_name = "output"
    ep_by_group_only = "annet.connectors.output"

    def _get_default(self) -> Type["OutputDriver"]:
        return OutputDriverBasic


output_driver_connector = _DriverConnector()


class OutputDriver(abc.ABC):
    @abc.abstractmethod
    def write_output(self, arg_out: FileOutOptions, items, query_result_count=1) -> None:
        pass

    @abc.abstractmethod
    def format_fails(self, fail, fqdns: Optional[Dict[int, str]] = None) -> Tuple[str, str]:
        pass

    @abc.abstractmethod
    def cfg_file_names(self, device: Device) -> List[str]:
        pass

    @abc.abstractmethod
    def entire_config_dest_path(self, device: Device, config_path: str) -> str:
        pass


class OutputDriverBasic(OutputDriver):
    def write_output(self, arg_out: FileOutOptions, items, query_result_count=1):
        """
        пишет результаты генерации в файл или директорию :dest
        :dest - это директория в случаях:
        - заканчивается на "/"
        - существует директория с таким именем
        - более одного устройства в результате запроса
        - есть entire-генераторы (определяется по типу первого результата, если устройство одно)
        """
        logger = get_logger()

        items_iter = iter(items)
        try:
            first_result = next(items_iter)
        except StopIteration:
            # нет результатов, ничего не пишем и не создаём
            return

        def _reassemble_items():
            yield first_result
            yield from items_iter

        dest = arg_out.dest
        suggest_dir = arg_out.dest_force_create_dir or os.sep in first_result[0]
        dir_mode = dir_or_file_output(dest, query_result_count, suggest_dir=suggest_dir)
        _reassembled_items = list(_reassemble_items())

        for output_no, (label, output, is_fail) in enumerate(_reassembled_items):
            writer = output if isinstance(output, OutputWriter) else OutputWriter(output)
            label = os.path.normpath(label)
            label_color = colorama.Back.RED if is_fail else colorama.Back.GREEN
            if dest is None:
                if hasattr(arg_out, "format") and arg_out.format == "json":
                    if output_no > 0:
                        sys.stdout.write(",")
                    elif output_no == 0:
                        sys.stdout.write("{")
                    sys.stdout.write('"%s": ' % label)
                    writer.write(sys.stdout)
                    if output_no == len(_reassembled_items) - 1:
                        sys.stdout.write("}")
                else:
                    if not arg_out.no_label:
                        print_label(label, back_color=label_color)
                    writer.write(sys.stdout)
            elif dir_mode:
                if arg_out.dest_force_create_dir and os.sep not in label:
                    label = os.path.join(label, BLACKBOX_FILENAME)

                if label.startswith(LABEL_NEW_PREFIX):
                    label = label[len(LABEL_NEW_PREFIX):]
                if label.startswith(os.sep):
                    # just in case.
                    label = label.lstrip(os.sep)

                if os.sep not in label:
                    # vendor config
                    label = os.path.basename(label)

                else:
                    # entire generated file
                    parts = label.split(os.sep)
                    hostname = parts[0]
                    label = os.sep.join(parts[1:])
                    if not arg_out.expand_path:
                        label = os.path.basename(label)
                    if query_result_count > 1 or arg_out.dest_force_create_dir:
                        label = os.path.join(hostname, label)
                file_dest = os.path.join(dest, "errors") if is_fail else dest

                out_file = os.path.normpath(os.path.join(file_dest, label))
                logger.info("writing '%s'", out_file)
                dirname = os.path.dirname(out_file)
                if not os.path.exists(dirname):
                    os.makedirs(dirname)
                with open(out_file, "w") as file:
                    writer.write(file)
            else:
                logger.info("writing '%s'", dest)
                with open(dest, "w") as file:
                    writer.write(file)

    def format_fails(self, fail, fqdns: Optional[Dict[int, str]] = None):
        ret = []
        fqdns = fqdns or {}
        for (assignment, exc) in fail.items():
            label = assignment
            if assignment in fqdns:
                label = fqdns[assignment]
            elif isinstance(assignment, tuple):
                label = assignment[0]
            else:
                ValueError("Failed to parse failed assignment %r" % assignment)
            ret.append((label, getattr(exc, "formatted_output", f"{repr(exc)} (formatted_output is absent)"), True))
        return ret

    def cfg_file_names(self, device: Device) -> list[str]:
        res = []
        if device.hostname:
            res.append(f"{device.hostname}.cfg")
        if device.id is not None and device.id != "":
            res.append(f"_id_{device.id}.cfg")
        if not res:
            raise RuntimeError("Neither hostname nor id is known for device")
        return res

    def entire_config_dest_path(self, device, config_path: str) -> str:
        """Формирует путь к конфигу в директории destname.

        Например, для устройства с hostname `my-device`:
        ```
        >>> device.entire_config_dest_path("/etc/frr/frr.conf")
        'my-device.cfg/etc/frr/frr.conf'
        >>>
        ```
        """

        # NOTE: с полученным `config_path` работаем через `posixpath`, а не через `os.path`, потому что
        #       entire-путь POSIX-специфичный; но в конце формируем путь через `os.path` для текущей платформы
        if not posixpath.abspath(config_path):
            raise RuntimeError(f"Want absolute config path, but relative received: {config_path}")

        cfg_files = self.cfg_file_names(device)

        parsed_config_path = urlparse(config_path)
        scheme, config_path = parsed_config_path.scheme or "file", parsed_config_path.path

        relative_config_path = posixpath.relpath(config_path, "/")
        if scheme != "file":
            host = parsed_config_path.hostname + f":{parsed_config_path.port}" if parsed_config_path.port else ""
            relative_config_path = os.path.join(host, relative_config_path)

        dest_config_path_parts = [cfg_files[0]] + relative_config_path.split(posixpath.sep)

        return os.path.join(*dest_config_path_parts)
